import pytest
import matplotlib

matplotlib.use('Agg')

pytestmark = pytest.mark.example

def test_ldc_example():
    from examples.ldc import main

    main(nx=4)

def test_ldc2_example():
    from examples.ldc2 import main

    main(nx=4)

def test_ldc3_example():
    from examples.ldc3 import main

    main(nx=4)

def test_ldc_3d_example():
    try:
        from examples.ldc_3d import main

        main(nx=4)
    except ImportError:
        pytest.skip("HYMLS not found")

def test_dhc_example():
    from examples.dhc import main

    main(nx=16)

def test_qg_example():
    from examples.qg import main

    main()

def test_amoc_example():
    from examples.amoc import main

    main(nx=4)



import os
import numpy
import pytest

from transiflow import utils
from transiflow import CrsMatrix
from transiflow import Discretization, CylindricalDiscretization

def create_coordinate_vector(nx):
    dx = 1 / (nx + 1)
    a = []
    for i in range(nx+3):
        a.append(-dx + dx * 1.2 ** i)
    return numpy.roll(a, -2)

def create_test_problem():
    nx = 13
    ny = 7
    nz = 5
    dim = 3
    dof = 5

    x = create_coordinate_vector(nx)
    y = create_coordinate_vector(ny)
    z = create_coordinate_vector(nz)

    parameters = {'Reynolds Number': 100}

    return (parameters, nx, ny, nz, dim, dof, x, y, z)

def test_uniform_grid():
    nx = 5
    dx = 1 / nx
    x = utils.create_uniform_coordinate_vector(0, 1, nx)
    for i in range(nx):
        assert x[i] == pytest.approx((i + 1) * dx)

    assert x[-1] == pytest.approx(0)
    assert x[0] > 0
    assert x[nx-1] == pytest.approx(1)

def test_shifted_uniform_grid():
    nx = 5
    dx = 3 / nx
    x = utils.create_uniform_coordinate_vector(1, 4, nx)
    for i in range(nx):
        assert x[i] == pytest.approx((i + 1) * dx + 1)

    assert x[-1] == pytest.approx(1)
    assert x[0] > 1
    assert x[nx-1] == pytest.approx(4)

def test_stretched_grid():
    nx = 5
    x = utils.create_stretched_coordinate_vector(0, 1, nx, 1.5)

    assert x[-1] == pytest.approx(0)
    assert x[0] > 0
    assert x[nx-1] == pytest.approx(1)

def test_u_xx():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.u_xx()

    for i in range(nx):
        dx = x[i] - x[i-1]
        dxp1 = x[i+1] - x[i]
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 0, 0, 0, 1, 1] == pytest.approx(1 / dx * dy * dz)
                assert atom[i, j, k, 0, 0, 2, 1, 1] == pytest.approx(1 / dxp1 * dy * dz)

def test_v_yy():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.v_yy()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            dy = y[j] - y[j-1]
            dyp1 = y[j+1] - y[j]
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 1, 1, 1, 0, 1] == pytest.approx(1 / dy * dx * dz)
                assert atom[i, j, k, 1, 1, 1, 2, 1] == pytest.approx(1 / dyp1 * dx * dz)

def test_w_zz():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.w_zz()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                dz = z[k] - z[k-1]
                dzp1 = z[k+1] - z[k]
                print(i, j, k)
                assert atom[i, j, k, 2, 2, 1, 1, 0] == pytest.approx(1 / dz * dy * dx)
                assert atom[i, j, k, 2, 2, 1, 1, 2] == pytest.approx(1 / dzp1 * dy * dx)

def test_u_yy():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.u_yy()

    for i in range(nx):
        dx = (x[i+1] - x[i-1]) / 2
        for j in range(ny):
            dy = (y[j] - y[j-2]) / 2
            dyp1 = (y[j+1] - y[j-1]) / 2
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 0, 0, 1, 0, 1] == pytest.approx(1 / dy * dx * dz)
                assert atom[i, j, k, 0, 0, 1, 2, 1] == pytest.approx(1 / dyp1 * dx * dz)

def test_v_xx():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.v_xx()

    for i in range(nx):
        dx = (x[i] - x[i-2]) / 2
        dxp1 = (x[i+1] - x[i-1]) / 2
        for j in range(ny):
            dy = (y[j+1] - y[j-1]) / 2
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 1, 1, 0, 1, 1] == pytest.approx(1 / dx * dy * dz)
                assert atom[i, j, k, 1, 1, 2, 1, 1] == pytest.approx(1 / dxp1 * dy * dz)

def test_w_yy():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.w_yy()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            dy = (y[j] - y[j-2]) / 2
            dyp1 = (y[j+1] - y[j-1]) / 2
            for k in range(nz):
                dz = (z[k+1] - z[k-1]) / 2
                print(i, j, k)
                assert atom[i, j, k, 2, 2, 1, 0, 1] == pytest.approx(1 / dy * dz * dx)
                assert atom[i, j, k, 2, 2, 1, 2, 1] == pytest.approx(1 / dyp1 * dz * dx)

def test_u_zz():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.u_zz()

    for i in range(nx):
        dx = (x[i+1] - x[i-1]) / 2
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                dz = (z[k] - z[k-2]) / 2
                dzp1 = (z[k+1] - z[k-1]) / 2
                print(i, j, k)
                assert atom[i, j, k, 0, 0, 1, 1, 0] == pytest.approx(1 / dz * dx * dy)
                assert atom[i, j, k, 0, 0, 1, 1, 2] == pytest.approx(1 / dzp1 * dx * dy)

def test_v_zz():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.v_zz()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            dy = (y[j+1] - y[j-1]) / 2
            for k in range(nz):
                dz = (z[k] - z[k-2]) / 2
                dzp1 = (z[k+1] - z[k-1]) / 2
                print(i, j, k)
                assert atom[i, j, k, 1, 1, 1, 1, 0] == pytest.approx(1 / dz * dy * dx)
                assert atom[i, j, k, 1, 1, 1, 1, 2] == pytest.approx(1 / dzp1 * dy * dx)

def test_w_xx():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.w_xx()

    for i in range(nx):
        dx = (x[i] - x[i-2]) / 2
        dxp1 = (x[i+1] - x[i-1]) / 2
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                dz = (z[k+1] - z[k-1]) / 2
                print(i, j, k)
                assert atom[i, j, k, 2, 2, 0, 1, 1] == pytest.approx(1 / dx * dz * dy)
                assert atom[i, j, k, 2, 2, 2, 1, 1] == pytest.approx(1 / dxp1 * dz * dy)

def test_T_xx():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.T_xx()

    for i in range(nx):
        dx = (x[i] - x[i-2]) / 2
        dxp1 = (x[i+1] - x[i-1]) / 2
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 4, 4, 0, 1, 1] == pytest.approx(1 / dx * dy * dz)
                assert atom[i, j, k, 4, 4, 2, 1, 1] == pytest.approx(1 / dxp1 * dy * dz)

def test_T_yy():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.T_yy()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            dy = (y[j] - y[j-2]) / 2
            dyp1 = (y[j+1] - y[j-1]) / 2
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 4, 4, 1, 0, 1] == pytest.approx(1 / dy * dx * dz)
                assert atom[i, j, k, 4, 4, 1, 2, 1] == pytest.approx(1 / dyp1 * dx * dz)

def test_T_zz():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.T_zz()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                dz = (z[k] - z[k-2]) / 2
                dzp1 = (z[k+1] - z[k-1]) / 2
                print(i, j, k)
                assert atom[i, j, k, 4, 4, 1, 1, 0] == pytest.approx(1 / dz * dy * dx)
                assert atom[i, j, k, 4, 4, 1, 1, 2] == pytest.approx(1 / dzp1 * dy * dx)

def test_p_x():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.p_x()

    for i in range(nx):
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 0, 3, 1, 1, 1] == pytest.approx(-dy * dz)
                assert atom[i, j, k, 0, 3, 2, 1, 1] == pytest.approx(dy * dz)

def test_p_y():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.p_y()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 1, 3, 1, 1, 1] == pytest.approx(-dx * dz)
                assert atom[i, j, k, 1, 3, 1, 2, 1] == pytest.approx(dx * dz)

def test_p_z():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.p_z()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                print(i, j, k)
                assert atom[i, j, k, 2, 3, 1, 1, 1] == pytest.approx(-dy * dx)
                assert atom[i, j, k, 2, 3, 1, 1, 2] == pytest.approx(dy * dx)

def test_u_x():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.u_x()

    for i in range(nx):
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 3, 0, 0, 1, 1] == pytest.approx(-dy * dz)
                assert atom[i, j, k, 3, 0, 1, 1, 1] == pytest.approx(dy * dz)

def test_u_y():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.v_y()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            for k in range(nz):
                dz = z[k] - z[k-1]
                print(i, j, k)
                assert atom[i, j, k, 3, 1, 1, 0, 1] == pytest.approx(-dx * dz)
                assert atom[i, j, k, 3, 1, 1, 1, 1] == pytest.approx(dx * dz)

def test_u_z():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    atom = discretization.w_z()

    for i in range(nx):
        dx = x[i] - x[i-1]
        for j in range(ny):
            dy = y[j] - y[j-1]
            for k in range(nz):
                print(i, j, k)
                assert atom[i, j, k, 3, 2, 1, 1, 0] == pytest.approx(-dy * dx)
                assert atom[i, j, k, 3, 2, 1, 1, 1] == pytest.approx(dy * dx)

def test_v_v():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    state_mtx = utils.create_padded_state_mtx(state, nx, ny, nz, dof, False, True, False)

    parameters['Problem Type'] = 'Taylor-Couette'
    discretization = CylindricalDiscretization(parameters, nx, ny, nz, dim, dof, x, y, z)

    atomJ = numpy.zeros([nx, ny, nz, dof, dof, 3, 3, 3])
    atomF = numpy.zeros([nx, ny, nz, dof, dof, 3, 3, 3])
    discretization.v_v(atomJ, atomF, state_mtx)

    averages_v = discretization.weighted_average_x(state_mtx[:, :, :, 1])
    averages_v = (averages_v[:, 0:ny, :] + averages_v[:, 1:ny+1, :]) / 2

    rhs = discretization.assemble_rhs(state, atomF)

    atom_value = numpy.zeros(1)
    for i in range(nx):
        for j in range(ny):
            for k in range(nz):
                print(i, j, k)
                Discretization._mass_x(atom_value, i, j, k, x, y, z)
                assert rhs[i * dof + j * nx * dof + k * nx * ny * dof] * x[i] == pytest.approx(
                    atom_value * averages_v[i, j, k+1] ** 2)

def read_matrix(fname):
    A = CrsMatrix([], [], [0])

    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'r') as f:
        rows = []
        idx = 0
        for i in f.readlines():
            r, c, v = [j.strip() for j in i.strip().split(' ') if j]
            r = int(r) - 1
            c = int(c) - 1
            v = float(v)
            rows.append(r)
            while r >= len(A.begA):
                A.begA.append(idx)
            A.jcoA.append(c)
            A.coA.append(v)
            idx += 1
        A.begA.append(idx)
        assert rows == sorted(rows)

    return A

def write_matrix(A, fname):
    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'w') as f:
        for i in range(len(A.begA)-1):
            for j in range(A.begA[i], A.begA[i+1]):
                f.write('%12d%12d %.16e\n' % (i+1, A.jcoA[j]+1, A.coA[j]))

def read_vector(fname):
    vec = numpy.array([])

    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'r') as f:
        for i in f.readlines():
            vec = numpy.append(vec, float(i.strip()))
    return vec

def write_vector(vec, fname):
    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'w') as f:
        for i in range(len(vec)):
            f.write('%.16e\n' % vec[i])

def assemble_jacobian(atom, nx, ny, nz, dof):
    row = 0
    idx = 0
    n = nx * ny * nz * dof
    coA = numpy.zeros(27*n)
    jcoA = numpy.zeros(27*n, dtype=int)
    begA = numpy.zeros(n+1, dtype=int)

    for k in range(nz):
        for j in range(ny):
            for i in range(nx):
                for d1 in range(dof):
                    for z in range(3):
                        for y in range(3):
                            for x in range(3):
                                for d2 in range(dof):
                                    if abs(atom[i, j, k, d1, d2, x, y, z]) > 1e-14:
                                        jcoA[idx] = row + (x-1) * dof + (y-1) * nx * dof + (z-1) * nx * ny * dof + d2 - d1
                                        coA[idx] = atom[i, j, k, d1, d2, x, y, z]
                                        idx += 1
                    row += 1
                    begA[row] = idx

    return CrsMatrix(coA, jcoA, begA)

def rotate_atom(atom, nx, ny, nz, dof):
    atom2 = numpy.zeros([nx, ny, nz, dof, dof, 3, 3, 3])
    for k in range(nz):
        for j in range(ny):
            for i in range(nx):
                for d1 in range(dof):
                    for z in range(3):
                        for y in range(3):
                            for x in range(3):
                                for d2 in range(dof):
                                    d3 = [2, 1, 0, 3][d1]
                                    d4 = [2, 1, 0, 3][d2]
                                    atom2[k, j, i, d3, d4, z, y, x] = atom[i, j, k, d1, d2, x, y, z]

    return atom2

def rotate_state(state, nx, ny, nz, dof):
    out = utils.create_state_mtx(state, nx, ny, nz, dof)
    state_mtx = utils.create_state_mtx(state, nx, ny, nz, dof)
    for k in range(nz):
        for j in range(ny):
            for i in range(nx):
                for d in range(dof):
                    d2 = [2, 1, 0, 3][d]
                    out[k, j, i, d2] = state_mtx[i, j, k, d]

    return utils.create_state_vec(out, nx, ny, nz, dof)

def test_rotate_lin():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100, 'Grid Stretching': True}

    parameters['X-max'] = 6
    parameters['Z-max'] = 1
    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom1 = discretization.linear_part()
    discretization.boundaries(atom1)

    parameters['X-max'] = 1
    parameters['Z-max'] = 6
    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom2 = discretization.linear_part()
    discretization.boundaries(atom2)

    atom2 = rotate_atom(atom2, nx, ny, nz, dof)

    tol = 1e-13
    assert numpy.all(abs(atom1 - atom2) < tol)

def test_rotate_bil():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100, 'Grid Stretching': True}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    parameters['X-max'] = 6
    parameters['Z-max'] = 1
    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atomJ1, atomF1 = discretization.nonlinear_part(state)
    discretization.boundaries(atomJ1)
    frc1 = discretization.boundaries(atomF1)

    state = rotate_state(state, nx, ny, nz, dof)

    parameters['X-max'] = 1
    parameters['Z-max'] = 6
    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atomJ2, atomF2 = discretization.nonlinear_part(state)
    discretization.boundaries(atomJ2)
    frc2 = discretization.boundaries(atomF2)

    atomJ2 = rotate_atom(atomJ2, nx, ny, nz, dof)
    atomF2 = rotate_atom(atomF2, nx, ny, nz, dof)
    frc2 = rotate_state(frc2, nx, ny, nz, dof)

    tol = 1e-13
    assert numpy.all(abs(atomJ1 - atomJ2) < tol)
    assert numpy.all(abs(atomF1 - atomF2) < tol)
    assert numpy.all(abs(frc1 - frc2) < tol)

def test_ldc_lin():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100}
    n = nx * ny * nz * dof

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom = discretization.linear_part()
    A = assemble_jacobian(atom, nx, ny, nz, dof)

    B = read_matrix('data/ldc_lin_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert B.begA[i+1] - B.begA[i] == A.begA[i+1] - A.begA[i]
        for j in range(B.begA[i], B.begA[i+1]):
            assert B.jcoA[j] == A.jcoA[j]
            assert B.coA[j] == A.coA[j]

def test_bous_lin():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 5
    parameters = {'Reynolds Number': 1, 'Rayleigh Number': 100, 'Prandtl Number': 100}
    n = nx * ny * nz * dof

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom = discretization.linear_part()
    A = assemble_jacobian(atom, nx, ny, nz, dof)

    B = read_matrix('data/bous_lin_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert B.begA[i+1] - B.begA[i] == A.begA[i+1] - A.begA[i]
        for j in range(B.begA[i], B.begA[i+1]):
            assert B.jcoA[j] == A.jcoA[j]
            assert B.coA[j] == A.coA[j]

def test_assemble_jacobian():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100}
    n = nx * ny * nz * dof

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom = discretization.linear_part()
    discretization.boundaries(atom)
    A = discretization.assemble_jacobian(atom)

    B = assemble_jacobian(atom, nx, ny, nz, dof)

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert B.begA[i+1] - B.begA[i] == A.begA[i+1] - A.begA[i]
        for j in range(B.begA[i], B.begA[i+1]):
            assert B.jcoA[j] == A.jcoA[j]
            assert B.coA[j] == A.coA[j]

def test_ldc_bnd():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100}
    n = nx * ny * nz * dof

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom = discretization.linear_part()
    discretization.boundaries(atom)
    A = discretization.assemble_jacobian(atom)

    B = read_matrix('data/ldc_bnd_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert B.begA[i+1] - B.begA[i] == A.begA[i+1] - A.begA[i]
        for j in range(B.begA[i], B.begA[i+1]):
            assert B.jcoA[j] == A.jcoA[j]
            assert B.coA[j] == A.coA[j]

def test_bous_bnd():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 5
    parameters = {'Reynolds Number': 1, 'Rayleigh Number': 100, 'Prandtl Number': 100, 'Problem Type': 'Rayleigh-Benard'}
    n = nx * ny * nz * dof

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom = discretization.linear_part()
    discretization.boundaries(atom)
    A = discretization.assemble_jacobian(atom)

    B = read_matrix('data/bous_bnd_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert B.begA[i+1] - B.begA[i] == A.begA[i+1] - A.begA[i]
        for j in range(B.begA[i], B.begA[i+1]):
            assert B.jcoA[j] == A.jcoA[j]
            assert B.coA[j] == pytest.approx(A.coA[j])

def test_manual_bous_bnd():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 5
    parameters = {'Rayleigh Number': 100, 'Prandtl Number': 100}
    n = nx * ny * nz * dof

    def boundaries(boundary_conditions, atom):
        boundary_conditions.heat_flux_east(atom, 0)
        boundary_conditions.heat_flux_west(atom, 0)
        boundary_conditions.no_slip_east(atom)
        boundary_conditions.no_slip_west(atom)

        boundary_conditions.heat_flux_north(atom, 0)
        boundary_conditions.heat_flux_south(atom, 0)
        boundary_conditions.no_slip_north(atom)
        boundary_conditions.no_slip_south(atom)

        Bi = 0
        boundary_conditions.heat_flux_top(atom, 0, Bi)
        boundary_conditions.temperature_bottom(atom, 0)
        boundary_conditions.free_slip_top(atom)
        boundary_conditions.no_slip_bottom(atom)

        return boundary_conditions.get_forcing()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof,
                                    boundary_conditions=boundaries)
    atom = discretization.linear_part()
    discretization.boundaries(atom)
    A = discretization.assemble_jacobian(atom)

    B = read_matrix('data/bous_bnd_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert B.begA[i+1] - B.begA[i] == A.begA[i+1] - A.begA[i]
        for j in range(B.begA[i], B.begA[i+1]):
            assert B.jcoA[j] == A.jcoA[j]
            assert B.coA[j] == pytest.approx(A.coA[j])

def test_ldc_bil():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom, atomF = discretization.nonlinear_part(state)
    discretization.boundaries(atom)
    A = discretization.assemble_jacobian(atom)

    B = read_matrix('data/ldc_bil_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        if i+1 >= len(B.begA):
            break

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        if A.begA[i+1] - A.begA[i] == 1 and A.coA[A.begA[i]] == -1:
            # Inserted boundary condition
            continue

        assert B.begA[i+1] - B.begA[i] == A.begA[i+1] - A.begA[i]
        for j in range(B.begA[i+1] - B.begA[i]):
            assert B.jcoA[B.begA[i] + j] == A.jcoA[A.begA[i] + j]
            assert B.coA[B.begA[i] + j] == A.coA[A.begA[i] + j]

def test_bous_bil():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 5
    parameters = {'Reynolds Number': 1, 'Rayleigh Number': 100, 'Prandtl Number': 100}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    atom, atomF = discretization.nonlinear_part(state)
    discretization.boundaries(atom)
    A = discretization.assemble_jacobian(atom)

    B = read_matrix('data/bous_bil_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        if i+1 >= len(B.begA):
            break

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        if A.begA[i+1] - A.begA[i] == 1 and A.coA[A.begA[i]] == -1:
            # Inserted boundary condition
            continue

        assert B.begA[i+1] - B.begA[i] == A.begA[i+1] - A.begA[i]
        for j in range(B.begA[i+1] - B.begA[i]):
            assert B.jcoA[B.begA[i] + j] == A.jcoA[A.begA[i] + j]
            assert B.coA[B.begA[i] + j] == A.coA[A.begA[i] + j]

def test_ldc():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    B = read_matrix('data/ldc_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/ldc_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])

def test_ldc_stretched():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100, 'Grid Stretching': True}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    B = read_matrix('data/ldc_stretched_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/ldc_stretched_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])

def test_ldc_2D():
    nx = 4
    ny = nx
    nz = 1
    dim = 3
    dof1 = 4
    parameters = {'Reynolds Number': 100}
    n = nx * ny * nz * dof1

    state1 = numpy.zeros(n)
    for i in range(n):
        state1[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof1)
    A1 = discretization.jacobian(state1)
    rhs1 = discretization.rhs(state1)

    dim = 2
    dof2 = 3
    n = nx * ny * nz * dof2

    state2 = numpy.zeros(n)
    for i in range(n):
        state2[i] = state1[i + (i + 1) // dof2]

    discretization = Discretization(parameters, nx, ny, nz, dim, dof2)
    A2 = discretization.jacobian(state2)
    rhs2 = discretization.rhs(state2)

    for i in range(n):
        print(i)

        i1 = i + (i + 1) // dof2

        print('Expected:')
        print(A1.jcoA[A1.begA[i1]:A1.begA[i1+1]])
        print(A1.coA[A1.begA[i1]:A1.begA[i1+1]])

        print('Got:')
        print(A2.jcoA[A2.begA[i]:A2.begA[i+1]])
        print(A2.coA[A2.begA[i]:A2.begA[i+1]])

        assert A1.begA[i1+1] - A1.begA[i1] == A2.begA[i+1] - A2.begA[i]
        for j in range(A2.begA[i+1] - A2.begA[i]):
            j1 = A1.begA[i1] + j
            j2 = A2.begA[i] + j
            assert A1.jcoA[j1] == A2.jcoA[j2] + (A2.jcoA[j2] + 1) // dof2
            assert A1.coA[j1] == pytest.approx(A2.coA[j2])

        assert rhs2[i] == pytest.approx(rhs1[i1])

def test_bous():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 5
    parameters = {'Reynolds Number': 1, 'Rayleigh Number': 100, 'Prandtl Number': 100,
                  'Problem Type': 'Rayleigh-Benard'}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    B = read_matrix('data/bous_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/bous_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])

def test_bous_stretched():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 5
    parameters = {'Reynolds Number': 1, 'Rayleigh Number': 100, 'Prandtl Number': 100,
                  'Problem Type': 'Rayleigh-Benard', 'Grid Stretching': True}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    B = read_matrix('data/bous_stretched_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/bous_stretched_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])

def test_bous_2D():
    nx = 4
    ny = nx
    nz = 1
    dim = 3
    dof1 = 5
    parameters = {'Reynolds Number': 1, 'Rayleigh Number': 100, 'Prandtl Number': 100,
                  'Problem Type': 'Rayleigh-Benard'}
    n = nx * ny * nz * dof1

    state1 = numpy.zeros(n)
    for i in range(n):
        state1[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof1)
    A1 = discretization.jacobian(state1)
    rhs1 = discretization.rhs(state1)

    dim = 2
    dof2 = 4
    n = nx * ny * nz * dof2

    state2 = numpy.zeros(n)
    for i in range(n):
        state2[i] = state1[i + (i + 2) // dof2]

    discretization = Discretization(parameters, nx, ny, nz, dim, dof2)
    A2 = discretization.jacobian(state2)
    rhs2 = discretization.rhs(state2)

    for i in range(n):
        print(i)

        i1 = i + (i + 2) // dof2

        print('Expected:')
        print(A1.jcoA[A1.begA[i1]:A1.begA[i1+1]])
        print(A1.coA[A1.begA[i1]:A1.begA[i1+1]])

        print('Got:')
        print(A2.jcoA[A2.begA[i]:A2.begA[i+1]])
        print(A2.coA[A2.begA[i]:A2.begA[i+1]])

        assert A1.begA[i1+1] - A1.begA[i1] == A2.begA[i+1] - A2.begA[i]
        for j in range(A2.begA[i+1] - A2.begA[i]):
            j1 = A1.begA[i1] + j
            j2 = A2.begA[i] + j
            assert A1.jcoA[j1] == A2.jcoA[j2] + (A2.jcoA[j2] + 2) // dof2
            assert A1.coA[j1] == pytest.approx(A2.coA[j2])

        assert rhs2[i] == pytest.approx(rhs1[i1])

def test_dhc():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 5
    parameters = {'Reynolds Number': 1, 'Rayleigh Number': 100, 'Prandtl Number': 100,
                  'Problem Type': 'Differentially Heated Cavity'}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    B = read_matrix('data/dhc_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/dhc_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])

def test_tc():
    nx = 4
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 1,
                  'Problem Type': 'Taylor-Couette'}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = CylindricalDiscretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    B = read_matrix('data/tc_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/tc_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])

def test_qg():
    nx = 4
    ny = nx
    nz = 1
    dim = 2
    dof = 4
    parameters = {'Reynolds Number': 16, 'Rossby Parameter': 1000,
                  'Wind Stress Parameter': 1000,
                  'Problem Type': 'Double Gyre'}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    B = read_matrix('data/qg_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/qg_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])

def test_amoc():
    nx = 4
    ny = nx
    nz = 1
    dim = 2
    dof = 5
    parameters = {'Reynolds Number': 16, 'Rayleigh Number': 4e4,
                  'Prandtl Number': 2.25, 'Lewis Number': 1,
                  'Temperature Forcing': 1, 'Freshwater Flux': 1,
                  'Problem Type': 'AMOC'}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    B = read_matrix('data/amoc_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/amoc_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])

def test_ldc8():
    nx = 8
    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {'Reynolds Number': 100}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    if not os.path.isfile('ldc_%sx%sx%s.txt' % (nx, ny, nz)):
        return

    B = read_matrix('data/ldc_%sx%sx%s.txt' % (nx, ny, nz))
    rhs_B = read_vector('data/ldc_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        print(i)

        print('Expected:')
        print(B.jcoA[B.begA[i]:B.begA[i+1]])
        print(B.coA[B.begA[i]:B.begA[i+1]])

        print('Got:')
        print(A.jcoA[A.begA[i]:A.begA[i+1]])
        print(A.coA[A.begA[i]:A.begA[i+1]])

        assert A.begA[i+1] - A.begA[i] == B.begA[i+1] - B.begA[i]
        for j in range(A.begA[i], A.begA[i+1]):
            assert A.jcoA[j] == B.jcoA[j]
            assert A.coA[j] == pytest.approx(B.coA[j])

        assert rhs_B[i] == pytest.approx(rhs[i])


if __name__ == '__main__':
    test_ldc8()



import numpy

from transiflow import utils
from transiflow import Discretization, CylindricalDiscretization

def create_coordinate_vector(nx):
    dx = 1 / (nx + 1)
    a = []
    for i in range(nx+3):
        a.append(-dx + dx * 1.2 ** i)
    return numpy.roll(a, -2)

def create_test_problem():
    nx = 13
    ny = 7
    nz = 5
    dim = 3
    dof = 5

    x = create_coordinate_vector(nx)
    y = create_coordinate_vector(ny)
    z = create_coordinate_vector(nz)

    parameters = {'Reynolds Number': 100}

    return (parameters, nx, ny, nz, dim, dof, x, y, z)

def create_random_state(nx, ny, nz, dim, dof):
    n = nx * ny * nz * dof
    x = numpy.random.random(n)
    x = utils.create_state_mtx(x, nx, ny, nz, dof)
    x[nx-1, :, :, 0] = 0
    x[:, ny-1, :, 1] = 0
    if dim > 2:
        x[:, :, nz-1, 2] = 0
    return utils.create_state_vec(x, nx, ny, nz, dof)

def create_test_problem_tc():
    nx = 13
    ny = 7
    nz = 5
    dim = 3
    dof = 4

    x = create_coordinate_vector(nx)
    y = create_coordinate_vector(ny)
    z = create_coordinate_vector(nz)

    parameters = {'Reynolds Number': 100, 'Problem Type': 'Taylor-Couette'}

    return (parameters, nx, ny, nz, dim, dof, x, y, z)

def create_random_state_tc(nx, ny, nz, dim, dof):
    n = nx * ny * nz * dof
    x = numpy.random.random(n)
    x = utils.create_state_mtx(x, nx, ny, nz, dof)
    x[nx-1, :, :, 0] = 0
    x[:, :, nz-1, 2] = 0
    return utils.create_state_vec(x, nx, ny, nz, dof)

def check_divfree(discretization, state):
    A = discretization.jacobian(state)
    x = A @ state
    for i in range(len(state)):
        if i % discretization.dof == discretization.dim:
            assert abs(x[i]) < 1e-13

def make_divfree(discretization, state):
    A = discretization.jacobian(state)
    p = numpy.zeros((A.n, A.n // discretization.dof))

    for i in range(A.n):
        if i % discretization.dof == discretization.dim:
            for j in range(A.begA[i], A.begA[i+1]):
                p[A.jcoA[j], i // discretization.dof] = A.coA[j]

    state -= p @ numpy.linalg.solve(p.conj().T @ p, p.conj().T @ state)

    check_divfree(discretization, state)

    return state

def create_divfree_state(discretization):
    nx = discretization.nx
    ny = discretization.ny
    nz = discretization.nz
    dim = discretization.dim
    dof = discretization.dof

    state = create_random_state(nx, ny, nz, dim, dof)
    return make_divfree(discretization, state)

def test_bilin():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    state = create_divfree_state(discretization)

    atomJ, atomF = discretization.nonlinear_part(state)
    discretization.boundaries(atomF)
    A = discretization.assemble_jacobian(atomF)

    for i in range(A.n):
        for j in range(A.begA[i], A.begA[i+1]):
            assert i != A.jcoA[j] or A.coA[j] == -1

def test_bilin_uniform():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    state = create_divfree_state(discretization)

    atomJ, atomF = discretization.nonlinear_part(state)
    discretization.boundaries(atomF)
    A = discretization.assemble_jacobian(atomF)

    for i in range(A.n):
        for j in range(A.begA[i], A.begA[i+1]):
            assert i != A.jcoA[j] or A.coA[j] == -1

def test_bilin_stretched():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    x = utils.create_stretched_coordinate_vector(0, 1, nx, 1.5)
    y = utils.create_stretched_coordinate_vector(0, 1, ny, 1.5)
    z = utils.create_stretched_coordinate_vector(0, 1, nz, 1.5)

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    state = create_divfree_state(discretization)

    atomJ, atomF = discretization.nonlinear_part(state)
    discretization.boundaries(atomF)
    A = discretization.assemble_jacobian(atomF)

    for i in range(A.n):
        for j in range(A.begA[i], A.begA[i+1]):
            assert i != A.jcoA[j] or A.coA[j] == -1

def test_jac_consistency():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    state = create_random_state(nx, ny, nz, dim, dof)
    pert = create_random_state(nx, ny, nz, dim, dof)

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    for i in range(3, 12):
        eps = 10 ** -i
        eps2 = max(eps, 10 ** (-14+i))
        rhs2 = discretization.rhs(state + eps * pert)
        assert numpy.linalg.norm((rhs2 - rhs) / eps - A @ pert) < eps2

def test_jac_consistency_uniform():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    state = create_random_state(nx, ny, nz, dim, dof)
    pert = create_random_state(nx, ny, nz, dim, dof)

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    for i in range(3, 12):
        eps = 10 ** -i
        eps2 = max(eps, 10 ** (-14+i))
        rhs2 = discretization.rhs(state + eps * pert)
        assert numpy.linalg.norm((rhs2 - rhs) / eps - A @ pert) < eps2

def test_jac_consistency_stretched():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    x = utils.create_stretched_coordinate_vector(0, 1, nx, 1.5)
    y = utils.create_stretched_coordinate_vector(0, 1, ny, 1.5)
    z = utils.create_stretched_coordinate_vector(0, 1, nz, 1.5)

    state = create_random_state(nx, ny, nz, dim, dof)
    pert = create_random_state(nx, ny, nz, dim, dof)

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y, z)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    for i in range(3, 12):
        eps = 10 ** -i
        eps2 = max(eps, 10 ** (-14+i))
        rhs2 = discretization.rhs(state + eps * pert)
        assert numpy.linalg.norm((rhs2 - rhs) / eps - A @ pert) < eps2

def test_jac_consistency_uniform_2d():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    nz = 1
    dim = 2
    dof = 3
    state = create_random_state(nx, ny, nz, dim, dof)
    pert = create_random_state(nx, ny, nz, dim, dof)

    discretization = Discretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    for i in range(3, 12):
        eps = 10 ** -i
        eps2 = max(eps, 10 ** (-14+i))
        rhs2 = discretization.rhs(state + eps * pert)
        assert numpy.linalg.norm((rhs2 - rhs) / eps - A @ pert) < eps2

def test_jac_consistency_stretched_2d():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem()

    x = utils.create_stretched_coordinate_vector(0, 1, nx, 1.5)
    y = utils.create_stretched_coordinate_vector(0, 1, ny, 1.5)

    nz = 1
    dim = 2
    dof = 3
    state = create_random_state(nx, ny, nz, dim, dof)
    pert = create_random_state(nx, ny, nz, dim, dof)

    discretization = Discretization(parameters, nx, ny, nz, dim, dof, x, y)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    for i in range(3, 12):
        eps = 10 ** -i
        eps2 = max(eps, 10 ** (-14+i))
        rhs2 = discretization.rhs(state + eps * pert)
        assert numpy.linalg.norm((rhs2 - rhs) / eps - A @ pert) < eps2

def test_jac_consistency_tc_uniform():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem_tc()

    state = create_random_state_tc(nx, ny, nz, dim, dof)
    pert = create_random_state_tc(nx, ny, nz, dim, dof)

    discretization = CylindricalDiscretization(parameters, nx, ny, nz, dim, dof)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    for i in range(3, 12):
        eps = 10 ** -i
        eps2 = 3 * max(eps, 10 ** (-14+i))
        rhs2 = discretization.rhs(state + eps * pert)
        assert numpy.linalg.norm((rhs2 - rhs) / eps - A @ pert) < eps2

def test_jac_consistency_tc_stretched():
    parameters, nx, ny, nz, dim, dof, x, y, z = create_test_problem_tc()

    x = utils.create_stretched_coordinate_vector(0, 1, nx, 1.5)
    z = utils.create_stretched_coordinate_vector(0, 1, nz, 1.5)

    state = create_random_state_tc(nx, ny, nz, dim, dof)
    pert = create_random_state_tc(nx, ny, nz, dim, dof)

    discretization = CylindricalDiscretization(parameters, nx, ny, nz, dim, dof, x, None, z)
    A = discretization.jacobian(state)
    rhs = discretization.rhs(state)

    for i in range(3, 12):
        eps = 10 ** -i
        eps2 = 3 * max(eps, 10 ** (-14+i))
        rhs2 = discretization.rhs(state + eps * pert)
        assert numpy.linalg.norm((rhs2 - rhs) / eps - A @ pert) < eps2



import numpy

from math import sqrt

def dot(x, y):
    '''Dot product between the vectors ``x`` and ``y`` that can be used for
    any backend'''
    try:
        return x.T.conj() @ y
    except AttributeError:
        return x.dot(y)

def norm(x):
    '''2-norm of the vector ``x`` that can be used for any backend'''
    if len(x.shape) > 1:
        ret = numpy.zeros(x.shape[1])
        for i in range(x.shape[1]):
            ret[i] = sqrt(abs(dot(x[:, i], x[:, i])))
        return ret

    return sqrt(abs(dot(x, x)))

def create_state_mtx(state, nx=None, ny=None, nz=None, dof=None, interface=None):
    '''Helper to create an (nx, ny, nz, dof) dimensional array out of a
    state vector that makes it easier to access the variables.

    Parameters
    ----------
    state : array_like
        State vector that you want to convert.
    nx : int
        Grid size in the x direction.
    ny : int
        Grid size in the y direction.
    nz : int, optional
        Grid size in the z direction.
    dof : int, optional
        Degrees of freedom of the problem associated with the state
        vector.
    interface : Interface, optional
        Sets ``nx, ny, nz, dof`` to the right values if supplied.

    Returns
    -------
    state_mtx : array_like
        An ``(nx, ny, nz, dof)`` dimensional array.

    '''
    if interface:
        nx = interface.nx
        ny = interface.ny
        nz = interface.nz
        dof = interface.dof

        state = interface.array_from_vector(state)

    state_mtx = numpy.zeros((nx, ny, nz, dof))
    for k, j, i, d in numpy.ndindex(nz, ny, nx, dof):
        state_mtx[i, j, k, d] = state[d + i * dof + j * dof * nx + k * dof * nx * ny]
    return state_mtx

def create_padded_state_mtx(state, nx=None, ny=None, nz=None, dof=None,
                            x_periodic=True, y_periodic=True, z_periodic=True,
                            interface=None):
    '''Helper to create an (nx+2, ny+2, nz+2, dof) dimensional array
    out of a state vector that makes it easier to access the
    variables. The value from the other side of the domain is padded
    to each border in case the domain is periodic.

    Parameters
    ----------
    state : array_like
        State vector that you want to convert.
    nx : int
        Grid size in the x direction.
    ny : int
        Grid size in the y direction.
    nz : int, optional
        Grid size in the z direction.
    dof : int, optional
        Degrees of freedom of the problem associated with the state
        vector.
    x_periodic : bool, optional
        Use periodic borders in the x direction.
    y_periodic : bool, optional
        Use periodic borders in the y direction.
    z_periodic : bool, optional
        Use periodic borders in the z direction.
    interface : Interface, optional
        Sets ``nx, ny, nz, dof, x_periodic, y_periodic, z_periodic`` to
        the right values if supplied.

    Returns
    -------
    state_mtx : array_like
        An ``(nx+2, ny+2, nz+2, dof)`` dimensional array.

    '''
    if interface:
        nx = interface.nx
        ny = interface.ny
        nz = interface.nz
        dof = interface.dof

        x_periodic = interface.discretization.x_periodic
        y_periodic = interface.discretization.y_periodic
        z_periodic = interface.discretization.z_periodic

        state = interface.array_from_vector(state)

    state_mtx = numpy.zeros((nx+2, ny+2, nz+2, dof))
    state_mtx[1:nx+1, 1:ny+1, 1:nz+1, :] = create_state_mtx(state, nx, ny, nz, dof)

    # Add extra borders for periodic boundary conditions
    if x_periodic:
        state_mtx[0, :, :, :] = state_mtx[nx, :, :, :]
        state_mtx[nx+1, :, :, :] = state_mtx[1, :, :, :]
    else:
        state_mtx[nx, :, :, 0] = 0

    if y_periodic:
        state_mtx[:, 0, :, :] = state_mtx[:, ny, :, :]
        state_mtx[:, ny+1, :, :] = state_mtx[:, 1, :, :]
    else:
        state_mtx[:, ny, :, 1] = 0

    if z_periodic:
        state_mtx[:, :, 0, :] = state_mtx[:, :, nz, :]
        state_mtx[:, :, nz+1, :] = state_mtx[:, :, 1, :]
    else:
        state_mtx[:, :, nz, 2] = 0

    return state_mtx

def create_state_vec(state_mtx, nx=None, ny=None, nz=None, dof=None, interface=None):
    '''Helper to create a state vector out of an array created with
    :meth:`create_state_mtx`.

    Parameters
    ----------
    state : array_like
        State vector that you want to convert.
    nx : int
        Grid size in the x direction.
    ny : int
        Grid size in the y direction.
    nz : int, optional
        Grid size in the z direction.
    dof : int, optional
        Degrees of freedom of the problem associated with the state
        vector.
    interface : Interface, optional
        Sets ``nx, ny, nz, dof`` to the right values if supplied.

    Returns
    -------
    state_mtx : array_like
        An ``nx * ny * nz * dof`` dimensional array.

    '''
    if interface:
        nx = interface.nx
        ny = interface.ny
        nz = interface.nz
        dof = interface.dof

    state = numpy.zeros(nx * ny * nz * dof)

    row = 0
    for k, j, i, d in numpy.ndindex(nz, ny, nx, dof):
        state[row] = state_mtx[i, j, k, d]
        row += 1
    return state

def create_uniform_coordinate_vector(start, end, nx):
    '''Create a uniformly distributed vector that can be used as
    coordinate vector in a Discretization.

    Parameters
    ----------
    start : int
        Start of the domain.
    end : int
        End of the domain.
    nx : int
        Amount of elements in this coordinate direction.

    Returns
    -------
    state : array_like
        An array of size ``nx + 3``. This includes padding for cells just
        outside of the domain.

    '''
    dx = (end - start) / nx
    x = start + numpy.arange(-1, nx + 2) * dx
    return numpy.roll(x, -2)

def create_stretched_coordinate_vector(start, end, nx, sigma):
    '''Create coordinated vector that can be used in a Discretization
    with more cells near the boundaries. This uses a `tanh` for stretching.

    Parameters
    ----------
    start : int
        Start of the domain.
    end : int
        End of the domain.
    nx : int
        Amount of elements in this coordinate direction.
    sigma : float
        Stretching factor.

    Returns
    -------
    state : array_like
        An array of size ``nx + 3``. This includes padding for cells just
        outside of the domain.

    '''
    x = create_uniform_coordinate_vector(0, 1, nx)
    x = 0.5 * (1 + numpy.tanh(2 * sigma * (x - 0.5)) / numpy.tanh(sigma))
    x = start + x * (end - start)

    # Make cells inside and outside of the boundary the same size
    # to make sure the boundary conditions are applied at the boundary
    dx = x[0] - x[-1]
    if start == 0:
        x[-2] = x[-1] - dx
    if end == 1:
        x[-3] = x[-4] + dx
    return x

def create_stretched_coordinate_vector2(start, end, nx, sigma):
    '''Create coordinated vector that can be used in a Discretization
    with more cells near the boundaries. This uses a `sin` for stretching.

    Parameters
    ----------
    start : int
        Start of the domain.
    end : int
        End of the domain.
    nx : int
        Amount of elements in this coordinate direction.
    sigma : float
        Stretching factor.

    Returns
    -------
    state : array_like
        An array of size ``nx + 3``. This includes padding for cells just
        outside of the domain.

    '''
    x = create_uniform_coordinate_vector(0, 1, nx)
    x = x - sigma * numpy.sin(2 * numpy.pi * x)
    x = start + x * (end - start)

    # Make cells inside and outside of the boundary the same size
    # to make sure the boundary conditions are applied at the boundary
    dx = x[0] - x[-1]
    if start == 0:
        x[-2] = x[-1] - dx
    if end == 1:
        x[-3] = x[-4] + dx
    return x

def compute_coordinate_vector_centers(vec):
    '''Compute the centers of cells in the direction of the supplied
    coordinate vector.

    Parameters
    ----------
    vec : array_like
        A coordinate vector

    Returns
    -------
    x : array_like
        A vector containing all of the cell centers.

    '''
    x = numpy.zeros(len(vec) - 1)
    for i in range(-1, len(vec) - 2):
        x[i] = (vec[i] + vec[i-1]) / 2

    return x

def compute_velocity_magnitude(state, interface, axis=2, position=None):
    '''Compute the velocity magnitude at the grid points in a plane.

    Parameters
    ----------
    state : array_like
        The state vector to extract the velocities from.
    interface : Interface
        Interface corresponding to the state vector.
    axis : int, optional
        Axis along which we take the center point, or a set position
        if provided. Not necessary in case the problem is 2D.
    position : float, optional
        Point along the provided axis at wich we take the plane.

    Returns
    -------
    x : array_like
        A 2D vector containing the velocity magnitudes.

    '''
    nx = interface.nx
    ny = interface.ny
    nz = interface.nz

    x = interface.x
    y = interface.y
    z = interface.z

    state_mtx = create_padded_state_mtx(state, interface=interface)

    # FIXME: This assumes zero or periodic boundaries
    if axis == 0:
        m = numpy.zeros((ny, nz))

        center = nx // 2 - 1
        if position:
            center = numpy.argmin(numpy.abs(x - position))

        print('Using center: %e at %d' % (x[center], center))

        for j, k in numpy.ndindex(ny, nz):
            u = get_u_value(state_mtx, center, j, k, interface)
            v = get_v_value(state_mtx, center, j, k, interface)
            w = get_w_value(state_mtx, center, j, k, interface)
            m[j, k] = sqrt(u * u + v * v + w * w)

        return m

    if axis == 1:
        m = numpy.zeros((nx, nz))

        center = ny // 2 - 1
        if position:
            center = numpy.argmin(numpy.abs(y - position))

        print('Using center: %e at %d' % (y[center], center))

        for i, k in numpy.ndindex(nx, nz):
            u = get_u_value(state_mtx, i, center, k, interface)
            v = get_v_value(state_mtx, i, center, k, interface)
            w = get_w_value(state_mtx, i, center, k, interface)
            m[i, k] = sqrt(u * u + v * v + w * w)

        return m

    m = numpy.zeros((nx, ny))

    center = nz // 2 - 1
    if position:
        center = numpy.argmin(numpy.abs(z - position))

    print('Using center: %e at %d' % (z[center], center))

    for i, j in numpy.ndindex(nx, ny):
        u = get_u_value(state_mtx, i, j, center, interface)
        v = get_v_value(state_mtx, i, j, center, interface)

        w = 0
        if interface.dim > 2:
            w = get_w_value(state_mtx, i, j, center, interface)

        m[i, j] = sqrt(u * u + v * v + w * w)

    return m

def compute_streamfunction(state, interface, axis=2):
    '''Compute the stream function at the grid points in a plane.

    Parameters
    ----------
    state : array_like
        The state vector to extract the velocities from.
    interface : Interface
        Interface corresponding to the state vector.
    axis : int, optional
        Axis along which we take the center point. Not necessary in
        case the problem is 2D.

    Returns
    -------
    x : array_like
        A 2D vector containing the stream function values.

    '''
    x = interface.x
    y = interface.y

    nx = interface.nx
    ny = interface.ny

    state_mtx = create_padded_state_mtx(state, interface=interface)

    center = interface.nz // 2 + 1
    u = state_mtx[1:, 1:, center, 0]
    v = state_mtx[1:, 1:, center, 1]

    if axis == 1:
        center = interface.ny // 2 + 1
        u = state_mtx[1:, center, 1:, 0]
        v = state_mtx[1:, center, 1:, 2]
        y = interface.z
        ny = interface.nz

    psiu = numpy.zeros((nx, ny))
    psiv = numpy.zeros((nx, ny))

    # Integration using the midpoint rule
    for i, j in numpy.ndindex(nx, ny):
        dx = x[i] - x[i-1]
        dy = y[j] - y[j-1]

        psiu[i, j] = v[i, j] * dx
        if i > 0:
            psiu[i, j] += psiu[i-1, j]

        psiv[i, j] = u[i, j] * dy
        if j > 0:
            psiv[i, j] += psiv[i, j-1]

    return (psiu - psiv) / 2

def compute_vorticity(state, interface, axis=2):
    '''Compute the vorticity at the grid points in a plane.

    Parameters
    ----------
    state : array_like
        The state vector to extract the velocities from.
    interface : Interface
        Interface corresponding to the state vector.
    axis : int, optional
        Axis along which we take the center point. Not necessary in
        case the problem is 2D.

    Returns
    -------
    x : array_like
        A 2D vector containing the vorticity values.

    '''
    x = interface.x
    y = interface.y

    nx = interface.nx
    ny = interface.ny

    state_mtx = create_padded_state_mtx(state, interface=interface)

    center = interface.nz // 2 + 1
    u = state_mtx[1:, 1:, center, 0]
    v = state_mtx[1:, 1:, center, 1]

    assert axis == 2

    zeta = numpy.zeros((nx, ny))

    # Integration using the midpoint rule
    for i, j in numpy.ndindex(nx, ny):
        dx = (x[i+1] - x[i-1]) / 2
        dy = (y[j+1] - y[j-1]) / 2

        if i < nx - 1:
            zeta[i, j] += v[i+1, j] / dx
            zeta[i, j] -= v[i, j] / dx

        if j < ny - 1:
            zeta[i, j] += u[i, j+1] / dy
            zeta[i, j] -= u[i, j] / dy

    return zeta

def compute_volume_averaged_kinetic_energy(state, interface):
    '''Compute the volume averaged kinetic energy at the grid points.

    Parameters
    ----------
    state : array_like
        The state vector to extract the velocities from.
    interface : Interface
        Interface corresponding to the state vector.

    Returns
    -------
    E : scalar
        The volume averaged kinetic energy.

    '''
    x = interface.x
    y = interface.y
    z = interface.z

    nx = interface.nx
    ny = interface.ny
    nz = interface.nz

    dim = interface.dim

    if nx <= 1:
        assert x[0] - x[-1] == 1

    if ny <= 1:
        assert y[0] - y[-1] == 1

    if nz <= 1:
        assert z[0] - z[-1] == 1

    state_mtx = create_padded_state_mtx(state, interface=interface)

    w = 0
    Ek = 0
    for i, j, k in numpy.ndindex(nx, ny, nz):
        dx = x[i] - x[i-1]
        dy = y[j] - y[j-1]
        dz = z[k] - z[k-1]

        u = (state_mtx[i+1, j+1, k+1, 0] + state_mtx[i, j+1, k+1, 0]) / 2
        v = (state_mtx[i+1, j+1, k+1, 1] + state_mtx[i+1, j, k+1, 1]) / 2
        if dim > 2:
            w = (state_mtx[i+1, j+1, k+1, 2] + state_mtx[i+1, j+1, k, 2]) / 2

        Ek += (u * u + v * v + w * w) * dx * dy * dz

    return Ek / 2

def get_u_value(state, i, j, k, interface):
    '''Get the value of u at a grid point.

    Parameters
    ----------
    state : array_like
        The state vector to extract the velocities from.
    i : int
        Index in the x direction.
    j : int
        Index in the y direction.
    k : int
        Index in the z direction.
    interface : Interface
        Interface corresponding to the state vector.

    Returns
    -------
    u : scalar
        Value at the grid point.

    '''
    if len(state.shape) < 4:
        state_mtx = create_padded_state_mtx(state, interface=interface)
    else:
        state_mtx = state

    y = interface.y
    dy1 = (y[j] - y[j-1]) / 2
    dy2 = (y[j+1] - y[j]) / 2

    u1 = (state_mtx[i+1, j+1, k+1, 0] * dy1 + state_mtx[i+1, j+2, k+1, 0] * dy2) / (dy1 + dy2)
    u2 = (state_mtx[i+1, j+1, k+2, 0] * dy1 + state_mtx[i+1, j+2, k+2, 0] * dy2) / (dy1 + dy2)

    z = interface.z
    dz1 = (z[k] - z[k-1]) / 2
    dz2 = (z[k+1] - z[k]) / 2

    return (u1 * dz1 + u2 * dz2) / (dz1 + dz2)

def get_v_value(state, i, j, k, interface):
    '''Get the value of v at a grid point.

    Parameters
    ----------
    state : array_like
        The state vector to extract the velocities from.
    i : int
        Index in the x direction.
    j : int
        Index in the y direction.
    k : int
        Index in the z direction.
    interface : Interface
        Interface corresponding to the state vector.

    Returns
    -------
    v : scalar
        Value at the grid point.

    '''
    if len(state.shape) < 4:
        state_mtx = create_padded_state_mtx(state, interface=interface)
    else:
        state_mtx = state

    x = interface.x
    dx1 = (x[i] - x[i-1]) / 2
    dx2 = (x[i+1] - x[i]) / 2

    v1 = (state_mtx[i+1, j+1, k+1, 1] * dx1 + state_mtx[i+2, j+1, k+1, 1] * dx2) / (dx1 + dx2)
    v2 = (state_mtx[i+1, j+1, k+2, 1] * dx1 + state_mtx[i+2, j+1, k+2, 1] * dx2) / (dx1 + dx2)

    z = interface.z
    dz1 = (z[k] - z[k-1]) / 2
    dz2 = (z[k+1] - z[k]) / 2

    return (v1 * dz1 + v2 * dz2) / (dz1 + dz2)

def get_w_value(state, i, j, k, interface):
    '''Get the value of w at a grid point.

    Parameters
    ----------
    state : array_like
        The state vector to extract the velocities from.
    i : int
        Index in the x direction.
    j : int
        Index in the y direction.
    k : int
        Index in the z direction.
    interface : Interface
        Interface corresponding to the state vector.

    Returns
    -------
    w : scalar
        Value at the grid point.

    '''
    if len(state.shape) < 4:
        state_mtx = create_padded_state_mtx(state, interface=interface)
    else:
        state_mtx = state

    x = interface.x
    dx1 = (x[i] - x[i-1]) / 2
    dx2 = (x[i+1] - x[i]) / 2

    w1 = (state_mtx[i+1, j+1, k+1, 2] * dx1 + state_mtx[i+2, j+1, k+1, 2] * dx2) / (dx1 + dx2)
    w2 = (state_mtx[i+1, j+2, k+1, 2] * dx1 + state_mtx[i+2, j+2, k+1, 2] * dx2) / (dx1 + dx2)

    y = interface.y
    dy1 = (y[j] - y[j-1]) / 2
    dy2 = (y[j+1] - y[j]) / 2

    return (w1 * dy1 + w2 * dy2) / (dy1 + dy2)



from transiflow import CrsMatrix

import numpy
import pytest

def get_test_matrix():
    coA = numpy.array([1, 2, 3, 4, 5], dtype=int)
    jcoA = numpy.array([0, 2, 1, 1, 2], dtype=int)
    begA = numpy.array([0, 2, 3, 5], dtype=int)
    return CrsMatrix(coA, jcoA, begA)

def get_random_test_matrix():
    m = 5
    n = 3
    coA = numpy.array(numpy.random.random(m * n), dtype=int)
    jcoA = numpy.repeat(numpy.array([numpy.arange(n)]), m)
    begA = numpy.arange(m * n + 1, step=n)
    return CrsMatrix(coA, jcoA, begA, m=m, n=n)

def test_getitem():
    A = get_test_matrix()

    assert A[0, 1] == 0
    assert A[1, 1] == 3

def test_getitem_range():
    A = get_test_matrix()
    B = A[0:2, 1:3]

    assert B[0, 0] == A[0, 1]
    assert B[0, 1] == A[0, 2]
    assert B[1, 0] == A[1, 1]
    assert B[1, 1] == A[1, 2]


def test_getitem_array():
    A = get_test_matrix()
    B = A[[0, 1], [0, 2]]

    assert B[0, 0] == A[0, 0]
    assert B[0, 1] == A[0, 2]
    assert B[1, 0] == A[1, 0]
    assert B[1, 1] == A[1, 2]

def test_setitem():
    A = get_test_matrix()

    B = CrsMatrix(m=A.m, n=A.n)
    B[0, 0] = 1
    B[0, 2] = 2
    B[1, 1] = 3
    B[2, 1] = 4
    B[2, 2] = 5
    B.assemble()

    for i in range(A.m):
        for j in range(A.n):
            assert A[i, j] == B[i, j]

def test_setitem_range():
    A = get_test_matrix()

    B = CrsMatrix(m=5, n=6)
    B[0:3, 1:4] = A
    B.assemble()

    assert A.begA[-1] == B.begA[-1]
    for i in range(A.m):
        for j in range(A.n):
            assert A[i, j] == B[i, j + 1]

def test_setitem_array():
    A = get_test_matrix()

    iidx = [1, 3, 4]
    jidx = [0, 2, 1]

    B = CrsMatrix(m=5, n=6)
    B[iidx, jidx] = A
    B.assemble()

    assert A.begA[-1] == B.begA[-1]
    for i in range(A.m):
        for j in range(A.n):
            assert A[i, j] == B[iidx[i], jidx[j]]

def test_setitem_to_array():
    A = get_test_matrix()
    A2 = A.to_dense()

    iidx = [1, 3, 4]
    jidx = [0, 2, 1]

    B = CrsMatrix(m=5, n=6)
    B[iidx, jidx] = A2
    B.assemble()

    assert A.begA[-1] == B.begA[-1]
    for i in range(A.m):
        for j in range(A.n):
            assert A[i, j] == B[iidx[i], jidx[j]]

def test_to_dense():
    A = get_test_matrix()
    B = A.to_dense()
    for i in range(A.m):
        for j in range(A.n):
            assert A[i, j] == B[i, j]

    A = get_random_test_matrix()
    B = A.coA.copy().reshape(A.m, A.n)
    for i in range(A.m):
        for j in range(A.n):
            assert A[i, j] == B[i, j]

def test_matvec():
    numpy.random.seed(1234)

    A = get_test_matrix()
    A2 = A.to_dense()

    x = numpy.random.random(A.n)
    b = A.matvec(x)
    b2 = numpy.matmul(A2, x)

    assert b == pytest.approx(b2)

    x = numpy.random.random((A.n, 4))
    b = A.matvec(x)
    b2 = numpy.matmul(A2, x)

    assert b == pytest.approx(b2)

def test_matmul():
    numpy.random.seed(1234)

    A = get_test_matrix()
    A2 = A.to_dense()

    x = numpy.random.random(A.n)
    b = A @ x
    b2 = A2 @ x

    assert b == pytest.approx(b2)

    x = numpy.random.random((A.n, 4))
    b = A @ x
    b2 = A2 @ x

    assert b == pytest.approx(b2)

def test_transpose():
    A = get_test_matrix()
    B = A.transpose()

    for i in range(A.m):
        for j in range(A.n):
            assert B[i, j] == A[j, i]

def test_add():
    A = get_test_matrix()
    B = A.to_dense()

    A2 = A + A
    B2 = B + B

    for i in range(A.m):
        for j in range(A.n):
            assert A2[i, j] == B2[i, j]
            assert A2[i, j] == A[i, j] + A[i, j]

def test_iadd():
    A = get_test_matrix()
    A2 = get_test_matrix()
    B = A.to_dense()

    A2 += A
    B2 = B + B

    for i in range(A.m):
        for j in range(A.n):
            assert A2[i, j] == B2[i, j]
            assert A2[i, j] == A[i, j] + A[i, j]

def test_sub():
    A = get_test_matrix()
    B = A.to_dense()

    A2 = A - A.transpose()
    B2 = B - B.T

    for i in range(A.m):
        for j in range(A.n):
            assert A2[i, j] == B2[i, j]
            assert A2[i, j] == A[i, j] - A[j, i]

def test_isub():
    A = get_test_matrix()
    A2 = A.transpose()
    B = A.to_dense()

    A2 -= A
    B2 = B.T - B

    for i in range(A.m):
        for j in range(A.n):
            assert A2[i, j] == B2[i, j]
            assert A2[i, j] == A[j, i] - A[i, j]



from .CrsMatrix import CrsMatrix
from .BoundaryConditions import BoundaryConditions
from .Discretization import Discretization
from .CylindricalDiscretization import CylindricalDiscretization
from .Continuation import Continuation
from .TimeIntegration import TimeIntegration

from .interface import create as Interface

__all__ = [
    'CrsMatrix', 'BoundaryConditions', 'Discretization', 'CylindricalDiscretization',
    'Continuation', 'TimeIntegration', 'Interface'
]



import numpy

from transiflow import utils
from transiflow import BoundaryConditions
from transiflow import CrsMatrix

class Discretization:
    r'''Finite volume discretization of the incompressible
    Navier-Stokes equations on a (possibly non-uniform) Arakawa
    C-grid. After discretization, the incompressible Navier-Stokes
    equations can be writen as a system of ordinary differential
    equations with algebraic constraints (DAE) of the form

    .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

    where $u$ is a state vector, $p$ is a set of parameters, $M(p)$ is
    a singular mass matrix, and $F(u, p)$ is the non-linear function
    including the forcing that is applied to the system.

    The variables in the state vector $u$ are ordered according to

    | ``[u, v, w, p, T, S, u, v, w, p, T, S, u, v, w, p, T, S, ...]``

    at positions

    | ``(0, 0, 0), (1, 0, 0), ..., (0, 1, 0), ..., (0, 0, 1), ...``

    Velocities ``u, v, w`` are staggered according to the C-grid
    definition, pressure ``p``, temperature ``T`` and salinity ``S``
    are defined in the centers of the grid cells. Variables are left
    out if they are not relevant to the problem. A 2D lid-driven
    cavity, for instance, only has ``[u, v, p]``.

    Parameters
    ----------
    parameters : dict
        Key-value pairs describing parameters of the model, for
        instance the Renolds number and the problem type. Possible
        values can be found in :ref:`problem definitions`.
    nx : int
        Grid size in the x direction.
    ny : int
        Grid size in the y direction.
    nz : int, optional
        Grid size in the z direction. 1 for 2-dimensional problems.
        This is the default.
    dim : int, optional
        Physical dimension of the problem. In case this is set to 2, w
        is not referenced in the state vector. The default is based on
        the value of nz.
    dof : int, optional
        Degrees of freedom for this problem. This should be set to dim
        plus 1 for each of pressure, temperature and salinity, if they
        are required for your problem. For example a 3D differentially
        heated cavity has dof = 3 + 1 + 1 = 5.
    x : array_like, optional
        Coordinate vector in the x direction.
    y : array_like, optional
        Coordinate vector in the y direction.
    z : array_like, optional
        Coordinate vector in the z direction.
    boundary_conditions : function, optional
        User-supplied function that implements the boundary
        conditions. It is called as ``boundary_conditions(bc, atom)``
        where ``bc`` is an instance of the :class:`.BoundaryConditions`
        class.

    Notes
    -----

    All discretizations are defined on atoms which for every grid cell
    define the contributions of neighbouring grid cells. In 3D this
    means 27 contributions from neighbouring grid cells are defined
    for every grid cell. For instance

    | ``atom[i, j, k, :, :, 1, 1, 1]``

    contains the contribution from the current cell at point
    ``(i, j, k)`` and

    | ``atom[i, j, k, :, :, 1, 0, 1]``

    contains the contribution from the cell south of the current one.
    A discretization of $u_{xx}$ on a uniform grid in 1D could be
    defined by

    | ``atom[i, j, k, :, :, 0, 1, 1] =  1 / dx``
    | ``atom[i, j, k, :, :, 1, 1, 1] = -2 / dx``
    | ``atom[i, j, k, :, :, 2, 1, 1] =  1 / dx``

    where we note that the mass matrix is also scaled by ``dx``.

    The remaining two indices, denoted by ``(:, :)`` above, contain
    the variable that is being used and the location of the equation
    that is being discretized. If we work in 2D, and we compute
    $p_x$ located in the first equation (as in the standard
    formulation of the incompressible Navier-Stokes equations), then
    the contribution is stored in

    | ``atom[:, :, :, 0, 2, :, :, :]``

    where the 0 comes from the first equation, 2 comes from the pressure.

    '''

    def __init__(self, parameters, nx, ny, nz=1, dim=None, dof=None,
                 x=None, y=None, z=None, boundary_conditions=None):
        self.parameters = parameters
        self.old_parameters = None

        self.nx = nx
        self.ny = ny
        self.nz = nz

        self.dim = dim
        if dim is None:
            self.dim = 3 if self.nz > 1 else 2

        self.dof = dof
        if dof is None:
            self.set_dof()

        self.x_periodic = False
        self.y_periodic = False
        self.z_periodic = False

        if self.nz == 1:
            self.z_periodic = True

        self.x = self.get_coordinate_vector(self.parameters.get('X-min', 0.0),
                                            self.parameters.get('X-max', 1.0),
                                            self.nx) if x is None else x
        self.y = self.get_coordinate_vector(self.parameters.get('Y-min', 0.0),
                                            self.parameters.get('Y-max', 1.0),
                                            self.ny) if y is None else y
        self.z = self.get_coordinate_vector(self.parameters.get('Z-min', 0.0),
                                            self.parameters.get('Z-max', 1.0),
                                            self.nz) if z is None else z

        self.atom = None

        self.boundary_conditions = boundary_conditions
        self._setup_boundary_conditions()

    def set_parameter(self, name, value):
        '''Set a parameter in ``self.parameters``. Changing a value in
        ``self.parameters`` will make us recompute the linear part of
        the equation.

        Parameters
        ----------
        name : str
            Name of the parameter
        value : Any
            Value of the parameter. Generally a floating point value

        '''

        if name in self.parameters and self.get_parameter(name) == value:
            return

        self.parameters[name] = value

    def get_parameter(self, name, default=0):
        '''Get a parameter from ``self.parameters``.

        Parameters
        ----------
        name : str
            Name of the parameter
        default : Any, optional
            Default return value if the parameter is not found in
            ``self.parameters``. The default value is 0.

        Returns
        -------
        The value of the parameter

        '''

        if name not in self.parameters:
            return default

        return self.parameters.get(name, default)

    def get_coordinate_vector(self, start, end, nx):
        '''Get a coordinate vector according to the set parameters.

        Parameters
        ----------
        start : float
            Start of the domain
        end : float
            End of the domain
        nx : int
            Amount of steps

        '''

        if self.parameters.get('Grid Stretching', False) or 'Grid Stretching Factor' in self.parameters.keys():
            if self.parameters.get('Grid Stretching Method', 'tanh') == 'sin':
                return utils.create_stretched_coordinate_vector2(
                    start, end, nx, self.parameters.get('Grid Stretching Factor', 0.1))

            return utils.create_stretched_coordinate_vector(
                start, end, nx, self.parameters.get('Grid Stretching Factor', 1.5))

        return utils.create_uniform_coordinate_vector(start, end, nx)

    def linear_part(self):
        '''Compute the linear part of the equation. Return a cached version if
        possible.

        :meta private:

        '''

        parameters = dict(self.parameters)
        if parameters != self.old_parameters:
            if self.dim == 2:
                self.atom = self._linear_part_2D()
            else:
                self.atom = self._linear_part_3D()

            self.old_parameters = parameters

        return self.atom

    def _linear_part_2D(self):
        '''Compute the linear part of the equation in case the domain
        is 2D. In case Re = 0 we instead compute the linear part for
        the Stokes problem.

        '''

        Re = self.get_parameter('Reynolds Number', 1.0)
        Ra = self.get_parameter('Rayleigh Number', 1.0)
        Pr = self.get_parameter('Prandtl Number', 1.0)
        Gr = self.get_parameter('Grashof Number', Ra / Pr)
        Le = self.get_parameter('Lewis Number', 1.0)

        if Re == 0:
            Re = 1

        if Gr == 0:
            Gr = 1 / Pr

        atom = 1 / (Re * numpy.sqrt(Gr)) * (self.u_xx() + self.u_yy()
                                            + self.v_xx() + self.v_yy()) \
            - (self.p_x() + self.p_y()) \
            + self.div()

        beta = self.get_parameter('Rossby Parameter')
        if beta:
            atom -= beta * self.coriolis()

        if self.dof > 3:
            atom += 1 / (Pr * numpy.sqrt(Gr)) * (self.T_xx() + self.T_yy())
            atom += self.forward_average_T_y()

        if self.dof > 4:
            atom += 1 / (Le * Pr * numpy.sqrt(Gr)) * (self.S_xx() + self.S_yy())
            atom -= self.forward_average_S_y()

        if self.problem_type_equals('Rayleigh-Benard Perturbation'):
            Bi = self.get_parameter('Biot Number')
            atom += Bi / (Bi + 1) * self.backward_average_v_y()

        return atom

    def _linear_part_3D(self):
        '''Compute the linear part of the equation in case the domain
        is 3D. In case Re = 0 we instead compute the linear part for
        the Stokes problem.

        '''

        Re = self.get_parameter('Reynolds Number', 1.0)
        Ra = self.get_parameter('Rayleigh Number', 1.0)
        Pr = self.get_parameter('Prandtl Number', 1.0)
        Gr = self.get_parameter('Grashof Number', Ra / Pr)
        Le = self.get_parameter('Lewis Number', 1.0)

        if Re == 0:
            Re = 1

        if Gr == 0:
            Gr = 1 / Pr

        atom = 1 / (Re * numpy.sqrt(Gr)) * (self.u_xx() + self.u_yy() + self.u_zz()
                                            + self.v_xx() + self.v_yy() + self.v_zz()
                                            + self.w_xx() + self.w_yy() + self.w_zz()) \
            - (self.p_x() + self.p_y() + self.p_z()) \
            + self.div()

        if self.dof > 4:
            atom += 1 / (Pr * numpy.sqrt(Gr)) * (self.T_xx() + self.T_yy() + self.T_zz())
            if self.nz > 1:
                atom += self.forward_average_T_z()
            else:
                atom += self.forward_average_T_y()

        if self.dof > 5:
            atom += 1 / (Le * Pr * numpy.sqrt(Gr)) * (self.S_xx() + self.S_yy() + self.S_zz())
            if self.nz > 1:
                atom -= self.forward_average_S_z()
            else:
                atom -= self.forward_average_S_y()

        if self.problem_type_equals('Rayleigh-Benard Perturbation'):
            Bi = self.get_parameter('Biot Number')
            if self.nz > 1:
                atom += Bi / (Bi + 1) * self.backward_average_w_z()
            else:
                atom += Bi / (Bi + 1) * self.backward_average_v_y()

        return atom

    def nonlinear_part(self, state):
        r'''Compute the nonlinear part of the equation. In case $\Re =
        0$ this does nothing.

        :meta private:

        '''

        state_mtx = utils.create_padded_state_mtx(state, self.nx, self.ny, self.nz, self.dof,
                                                  self.x_periodic, self.y_periodic, self.z_periodic)

        atomJ = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        atomF = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))

        Re = self.get_parameter('Reynolds Number')
        if Re == 0 and not self.dof > self.dim + 1:
            return (atomJ, atomF)

        self.u_u_x(atomJ, atomF, state_mtx)
        self.u_v_x(atomJ, atomF, state_mtx)
        self.v_u_y(atomJ, atomF, state_mtx)
        self.v_v_y(atomJ, atomF, state_mtx)

        if self.dim > 2:
            self.u_w_x(atomJ, atomF, state_mtx)
            self.v_w_y(atomJ, atomF, state_mtx)
            self.w_u_z(atomJ, atomF, state_mtx)
            self.w_v_z(atomJ, atomF, state_mtx)
            self.w_w_z(atomJ, atomF, state_mtx)

        if self.dof > self.dim + 1:
            self.u_T_x(atomJ, atomF, state_mtx)
            self.v_T_y(atomJ, atomF, state_mtx)

            if self.dim > 2:
                self.w_T_z(atomJ, atomF, state_mtx)

        if self.dof > self.dim + 2:
            self.u_S_x(atomJ, atomF, state_mtx)
            self.v_S_y(atomJ, atomF, state_mtx)

            if self.dim > 2:
                self.w_S_z(atomJ, atomF, state_mtx)

        atomJ += atomF

        return (atomJ, atomF)

    def rhs(self, state):
        r'''Compute the right-hand side of the DAE. That is the
        right-hand side $F(u, p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $F(u, p)$

        Returns
        -------
        rhs : array_like
            The value of $F(u, p)$

        '''

        atomJ, atomF = self.nonlinear_part(state)
        atomF += self.linear_part()

        frc = self.boundaries(atomF)

        return self.assemble_rhs(state, atomF) + frc

    def jacobian(self, state):
        r'''Compute the Jacobian matrix $J(u, p)$ of the right-hand
        side of the DAE. That is the Jacobian matrix of $F(u, p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $J(u, p)$

        Returns
        -------
        jac : CrsMatrix
            The matrix $J(u, p)$ in CSR format

        '''

        atomJ, atomF = self.nonlinear_part(state)
        atomJ += self.linear_part()

        self.boundaries(atomJ)

        return self.assemble_jacobian(atomJ)

    def mass_matrix(self):
        r'''Compute the mass matrix of the DAE. That is the mass
        matrix $M(p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        Returns
        -------
        mass : CrsMatrix
            The matrix $M(p)$ in CSR format

        '''

        atom = self.mass_x() + self.mass_y()
        if self.dim == 3:
            atom += self.mass_z()
        if self.dof > self.dim + 1:
            atom += self.mass_T()
        if self.dof > self.dim + 2:
            atom += self.mass_S()
        return self.assemble_mass_matrix(atom)

    def assemble_rhs(self, state, atom):
        '''Assemble the right-hand side. Optimized version of

        .. code-block:: Python

            for k in range(nz):
                for j in range(ny):
                    for i in range(nx):
                        for d1 in range(dof):
                            for z in range(3):
                                for y in range(3):
                                    for x in range(3):
                                        for d2 in range(dof):
                                            if abs(atom[i, j, k, d1, d2, x, y, z]) > 1e-14:
                                                offset = row + (x-1) * dof + (y-1) * nx * dof + \
                                                               (z-1) * nx * ny * dof + d2 - d1
                                                out[row] -= atom[i, j, k, d1, d2, x, y, z] * state[offset]
                            row += 1

        :meta private:

        '''

        # Put the state in shifted matrix form
        state_mtx = utils.create_padded_state_mtx(state, self.nx, self.ny, self.nz, self.dof,
                                                  self.x_periodic, self.y_periodic, self.z_periodic)

        # Add up all contributions without iterating over the domain
        out_mtx = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        for k, j, i, d1, d2 in numpy.ndindex(3, 3, 3, self.dof, self.dof):
            out_mtx[:, :, :, d1] += atom[:, :, :, d1, d2, i, j, k] \
                * state_mtx[i:(i+self.nx), j:(j+self.ny), k:(k+self.nz), d2]

        return utils.create_state_vec(out_mtx, self.nx, self.ny, self.nz, self.dof)

    def assemble_jacobian(self, atom):
        '''Assemble the Jacobian. Optimized version of

        .. code-block:: Python

            for k in range(nz):
                for j in range(ny):
                    for i in range(nx):
                        for d1 in range(dof):
                            for z in range(3):
                                for y in range(3):
                                    for x in range(3):
                                        for d2 in range(dof):
                                            if abs(atom[i, j, k, d1, d2, x, y, z]) > 1e-14:
                                               jcoA[idx] = row + (x-1) * dof + (y-1) * nx * dof + \
                                                                 (z-1) * nx * ny * dof + d2 - d1
                                               coA[idx] = atom[i, j, k, d1, d2, x, y, z]
                                               idx += 1
                            row += 1
                            begA[row] = idx

        :meta private:

        '''

        row = 0
        idx = 0
        n = self.nx * self.ny * self.nz * self.dof
        coA = numpy.zeros(27*n)
        jcoA = numpy.zeros(27*n, dtype=int)
        begA = numpy.zeros(n+1, dtype=int)

        # Check where values are nonzero in the atoms
        configs = []
        for d2, x, y, z in numpy.ndindex(self.dof, 3, 3, 3):
            if numpy.any(atom[:, :, :, :, d2, x, y, z]):
                configs.append((d2, x, y, z))

        # Iterate only over configurations with values in there
        for k, j, i, d1 in numpy.ndindex(self.nz, self.ny, self.nx, self.dof):
            for d2, x, y, z in configs:
                if abs(atom[i, j, k, d1, d2, x, y, z]) > 1e-14:
                    jcoA[idx] = ((i + x - 1) % self.nx) * self.dof \
                        + ((j + y - 1) % self.ny) * self.nx * self.dof + \
                        + ((k + z - 1) % self.nz) * self.nx * self.ny * self.dof + d2
                    coA[idx] = atom[i, j, k, d1, d2, x, y, z]
                    idx += 1
            row += 1
            begA[row] = idx

        return CrsMatrix(coA, jcoA, begA)

    def assemble_mass_matrix(self, atom):
        '''Assemble the mass matrix.

        :meta private:

        '''

        row = 0
        idx = 0
        n = self.nx * self.ny * self.nz * self.dof
        coA = numpy.zeros(n)
        jcoA = numpy.zeros(n, dtype=int)
        begA = numpy.zeros(n+1, dtype=int)

        for k, j, i, d in numpy.ndindex(self.nz, self.ny, self.nx, self.dof):
            if abs(atom[i, j, k, d]) > 1e-14:
                jcoA[idx] = (i + (j + k * self.ny) * self.nx) * self.dof + d
                coA[idx] = atom[i, j, k, d]
                idx += 1
            row += 1
            begA[row] = idx

        return CrsMatrix(coA, jcoA, begA)

    def problem_type_equals(self, second):
        '''Test against the problem type in ``self.parameterlist``.

        :meta private:

        '''
        first = self.get_parameter('Problem Type', 'Lid-driven Cavity')
        return first.lower() == second.lower()

    def set_dof(self):
        '''Set ``self.dof`` based on the problem type.

        :meta private:

        '''

        self.dof = self.dim + 1
        if self.problem_type_equals('Rayleigh-Benard') \
           or self.problem_type_equals('Rayleigh-Benard Perturbation'):
            self.dof = self.dim + 2
        elif self.problem_type_equals('Differentially Heated Cavity'):
            self.dof = self.dim + 2
        elif self.problem_type_equals('AMOC'):
            self.dof = self.dim + 3

    # Boundary conditions

    def _lid_driven_cavity(self, boundary_conditions, atom):
        '''Boundary conditions for the lid-driven cavity'''
        v = self.get_parameter('Lid Velocity', 1)
        boundary_conditions.no_slip_east(atom)
        boundary_conditions.no_slip_west(atom)

        boundary_conditions.no_slip_south(atom)
        if self.dim == 2 or self.nz <= 1:
            boundary_conditions.moving_lid_north(atom, v)
            return boundary_conditions.get_forcing()

        boundary_conditions.no_slip_north(atom)

        boundary_conditions.no_slip_bottom(atom)
        boundary_conditions.moving_lid_top(atom, v)

        return boundary_conditions.get_forcing()

    def _rayleigh_benard(self, boundary_conditions, atom):
        '''Boundary conditions for the Rayleigh-Benard problem'''
        asym = self.get_parameter('Asymmetry Parameter')
        boundary_conditions.heat_flux_east(atom, asym)
        boundary_conditions.heat_flux_west(atom, 0)
        boundary_conditions.no_slip_east(atom)
        boundary_conditions.no_slip_west(atom)

        Bi = self.get_parameter('Biot Number')
        bottom_temperature = 1 if self.problem_type_equals('Rayleigh-Benard') else 0

        if self.dim == 2 or self.nz <= 1:
            boundary_conditions.heat_flux_north(atom, 0, Bi)
            boundary_conditions.temperature_south(atom, bottom_temperature)
            boundary_conditions.free_slip_north(atom)
            boundary_conditions.no_slip_south(atom)
            return boundary_conditions.get_forcing()

        boundary_conditions.heat_flux_north(atom, 0)
        boundary_conditions.heat_flux_south(atom, 0)
        boundary_conditions.no_slip_north(atom)
        boundary_conditions.no_slip_south(atom)

        boundary_conditions.heat_flux_top(atom, 0, Bi)
        boundary_conditions.temperature_bottom(atom, bottom_temperature)
        boundary_conditions.free_slip_top(atom)
        boundary_conditions.no_slip_bottom(atom)

        return boundary_conditions.get_forcing()

    def _differentially_heated_cavity(self, boundary_conditions, atom):
        '''Boundary conditions for the differentially heated cavity'''
        boundary_conditions.temperature_east(atom, -1/2)
        boundary_conditions.temperature_west(atom, 1/2)
        boundary_conditions.no_slip_east(atom)
        boundary_conditions.no_slip_west(atom)

        boundary_conditions.heat_flux_north(atom, 0)
        boundary_conditions.heat_flux_south(atom, 0)
        boundary_conditions.no_slip_north(atom)
        boundary_conditions.no_slip_south(atom)

        if self.dim > 2 and self.nz > 1:
            boundary_conditions.heat_flux_top(atom, 0)
            boundary_conditions.heat_flux_bottom(atom, 0)
            boundary_conditions.no_slip_top(atom)
            boundary_conditions.no_slip_bottom(atom)

        return boundary_conditions.get_forcing()

    def _double_gyre(self, boundary_conditions, atom):
        '''Boundary conditions for the double-gyre QG problem'''
        frc = self.wind_stress()

        boundary_conditions.no_slip_east(atom)
        boundary_conditions.no_slip_west(atom)

        boundary_conditions.free_slip_north(atom)
        boundary_conditions.free_slip_south(atom)

        return frc

    def _amoc(self, boundary_conditions, atom):
        '''Boundary conditions for the 2D AMOC'''
        boundary_conditions.heat_flux_east(atom, 0)
        boundary_conditions.heat_flux_west(atom, 0)
        boundary_conditions.salinity_flux_east(atom, 0)
        boundary_conditions.salinity_flux_west(atom, 0)
        boundary_conditions.free_slip_east(atom)
        boundary_conditions.free_slip_west(atom)

        boundary_conditions.heat_flux_south(atom, 0)
        boundary_conditions.salinity_flux_south(atom, 0)
        boundary_conditions.free_slip_south(atom)

        x = utils.compute_coordinate_vector_centers(self.x)

        theta = self.get_parameter('Temperature Forcing')
        asym = self.get_parameter('Asymmetry Parameter')
        A = self.parameters.get('X-max', 1.0)

        T_S = numpy.zeros((self.nx + 2, self.nz + 2))
        T_S[:, 0] = 1 / 2 * ((1 - asym) * numpy.cos(2 * numpy.pi * (x / A - 1 / 2))
                             + asym * numpy.cos(numpy.pi * x / A) + 1)
        boundary_conditions.temperature_north(atom, theta * T_S)

        sigma = self.get_parameter('Freshwater Flux')
        p = 2

        Q_S = numpy.zeros((self.nx + 2, self.nz + 2))
        Q_S[:, 0] = 3 * numpy.cos(p * numpy.pi * (x / A - 1 / 2)) - 6 / (p * numpy.pi) * numpy.sin(p * numpy.pi / 2)
        boundary_conditions.salinity_flux_north(atom, sigma * Q_S)

        boundary_conditions.free_slip_north(atom)

        # Fix one salinity value
        row = self.dim + 2
        for k, j, i, d, z, y, x in numpy.ndindex(self.nz, self.ny, self.nx, self.dof, 3, 3, 3):
            if ((i + x - 1) % self.nx) * self.dof \
               + ((j + y - 1) % self.ny) * self.nx * self.dof + \
               + ((k + z - 1) % self.nz) * self.nx * self.ny * self.dof + self.dim + 2 == row:
                atom[i, j, k, d, self.dim+2, x, y, z] = 0

        atom[0, 0, 0, self.dim+2, self.dim+2, 1, 1, 1] = -1

        frc = boundary_conditions.get_forcing()
        frc[row] = 0
        return frc

    def boundaries(self, atom):
        '''Apply the boundary conditions specified in
        ``self.boundary_conditions``.

        :meta private:

        '''
        boundary_conditions = BoundaryConditions(
            self.nx, self.ny, self.nz, self.dim, self.dof, self.x, self.y, self.z)

        return self.boundary_conditions(boundary_conditions, atom)

    def _setup_boundary_conditions(self):

        '''Setup boundary conditions for the currently defined problem type.'''
        if self.boundary_conditions:
            return
        elif self.problem_type_equals('Lid-driven Cavity'):
            self.boundary_conditions = self._lid_driven_cavity
        elif (self.problem_type_equals('Rayleigh-Benard')
              or self.problem_type_equals('Rayleigh-Benard Perturbation')):
            self.boundary_conditions = self._rayleigh_benard
        elif self.problem_type_equals('Differentially Heated Cavity'):
            self.boundary_conditions = self._differentially_heated_cavity
        elif self.problem_type_equals('Double Gyre'):
            self.boundary_conditions = self._double_gyre
        elif self.problem_type_equals('AMOC'):
            self.boundary_conditions = self._amoc
        else:
            raise Exception('Invalid problem type %s' % self.get_parameter('Problem Type'))

    # Below are all of the discretizations of separate parts of
    # equations that we can solve using FVM. This takes into account
    # non-uniform grids. New discretizations such as derivatives have
    # to be implemented in a similar way.

    @staticmethod
    def _u_xx(atom, i, j, k, x, y, z):
        # distance between u[i] and u[i-1]
        dx = x[i] - x[i-1]
        # distance between u[i+1] and u[i]
        dxp1 = x[i+1] - x[i]
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # second order finite difference
        atom[0] = 1 / dx * dy * dz
        atom[2] = 1 / dxp1 * dy * dz
        atom[1] = -atom[0] - atom[2]

    def u_xx(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_xx(atom[i, j, k, 0, 0, :, 1, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def v_yy(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_xx(atom[i, j, k, 1, 1, 1, :, 1], j, i, k, self.y, self.x, self.z)
        return atom

    def w_zz(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_xx(atom[i, j, k, 2, 2, 1, 1, :], k, j, i, self.z, self.y, self.x)
        return atom

    @staticmethod
    def _u_yy(atom, i, j, k, x, y, z):
        # distance between u[j] and u[j-1]
        dy = (y[j] - y[j-2]) / 2
        # distance between u[j+1] and u[j]
        dyp1 = (y[j+1] - y[j-1]) / 2
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # second order finite difference
        atom[0] = 1 / dy * dx * dz
        atom[2] = 1 / dyp1 * dx * dz
        atom[1] = -atom[0] - atom[2]

    def u_yy(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_yy(atom[i, j, k, 0, 0, 1, :, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def v_xx(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_yy(atom[i, j, k, 1, 1, :, 1, 1], j, i, k, self.y, self.x, self.z)
        return atom

    def w_yy(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_yy(atom[i, j, k, 2, 2, 1, :, 1], k, j, i, self.z, self.y, self.x)
        return atom

    @staticmethod
    def _u_zz(atom, i, j, k, x, y, z):
        # distance between u[k] and u[k-1]
        dz = (z[k] - z[k-2]) / 2
        # distance between u[k+1] and u[k]
        dzp1 = (z[k+1] - z[k-1]) / 2
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2
        # volume size in the y direction
        dy = y[j] - y[j-1]

        # second order finite difference
        atom[0] = 1 / dz * dx * dy
        atom[2] = 1 / dzp1 * dx * dy
        atom[1] = -atom[0] - atom[2]

    def u_zz(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_zz(atom[i, j, k, 0, 0, 1, 1, :], i, j, k, self.x, self.y, self.z)
        return atom

    def v_zz(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_zz(atom[i, j, k, 1, 1, 1, 1, :], j, i, k, self.y, self.x, self.z)
        return atom

    def w_xx(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_zz(atom[i, j, k, 2, 2, :, 1, 1], k, j, i, self.z, self.y, self.x)
        return atom

    @staticmethod
    def _C_xx(atom, i, j, k, x, y, z):
        # distance between T[i] and T[i-1]
        dx = (x[i] - x[i-2]) / 2
        # distance between T[i+1] and T[i]
        dxp1 = (x[i+1] - x[i-1]) / 2
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # second order finite difference
        atom[0] = 1 / dx * dy * dz
        atom[2] = 1 / dxp1 * dy * dz
        atom[1] = -atom[0] - atom[2]

    def C_xx(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._C_xx(atom[i, j, k, var, var, :, 1, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def C_yy(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._C_xx(atom[i, j, k, var, var, 1, :, 1], j, i, k, self.y, self.x, self.z)
        return atom

    def C_zz(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._C_xx(atom[i, j, k, var, var, 1, 1, :], k, j, i, self.z, self.y, self.x)
        return atom

    def T_xx(self):
        ''':meta private:'''
        return self.C_xx(self.dim + 1)

    def T_yy(self):
        ''':meta private:'''
        return self.C_yy(self.dim + 1)

    def T_zz(self):
        ''':meta private:'''
        return self.C_zz(self.dim + 1)

    def S_xx(self):
        ''':meta private:'''
        return self.C_xx(self.dim + 2)

    def S_yy(self):
        ''':meta private:'''
        return self.C_yy(self.dim + 2)

    def S_zz(self):
        ''':meta private:'''
        return self.C_zz(self.dim + 2)

    @staticmethod
    def _forward_u_x(atom, i, j, k, x, y, z):
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # forward difference
        atom[2] = dy * dz
        atom[1] = -atom[2]

    def C_x(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._forward_u_x(atom[i, j, k, 0, var, :, 1, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def C_y(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._forward_u_x(atom[i, j, k, 1, var, 1, :, 1], j, i, k, self.y, self.x, self.z)
        return atom

    def C_z(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._forward_u_x(atom[i, j, k, 2, var, 1, 1, :], k, j, i, self.z, self.y, self.x)
        return atom

    def p_x(self):
        ''':meta private:'''
        return self.C_x(self.dim)

    def p_y(self):
        ''':meta private:'''
        return self.C_y(self.dim)

    def p_z(self):
        ''':meta private:'''
        return self.C_z(self.dim)

    @staticmethod
    def _backward_u_x(atom, i, j, k, x, y, z):
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # backward difference
        atom[1] = dy * dz
        atom[0] = -atom[1]

    def u_x(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_x(atom[i, j, k, self.dim, 0, :, 1, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def v_y(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_x(atom[i, j, k, self.dim, 1, 1, :, 1], j, i, k, self.y, self.x, self.z)
        return atom

    def w_z(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_x(atom[i, j, k, self.dim, 2, 1, 1, :], k, j, i, self.z, self.y, self.x)
        return atom

    @staticmethod
    def _backward_u_y(atom, i, j, k, x, y, z):
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # backward difference
        atom[1] = dx * dz
        atom[0] = -atom[1]

    @staticmethod
    def _backward_u_z(atom, i, j, k, x, y, z):
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2
        # volume size in the y direction
        dy = y[j] - y[j-1]

        # backward difference
        atom[1] = dx * dy
        atom[0] = -atom[1]

    def div(self):
        ''':meta private:'''
        if self.dim == 2:
            return self.u_x() + self.v_y()
        return self.u_x() + self.v_y() + self.w_z()

    @staticmethod
    def _forward_average_x(atom, i, j, k, x, y, z):
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # forward average
        atom[1] = dx * dy * dz / 2
        atom[2] = atom[1]

    def forward_average_C_y(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._forward_average_x(atom[i, j, k, 1, var, 1, :, 1], j, i, k, self.y, self.x, self.z)
        return atom

    def forward_average_C_z(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._forward_average_x(atom[i, j, k, 2, var, 1, 1, :], k, j, i, self.z, self.y, self.x)
        return atom

    def forward_average_T_y(self):
        ''':meta private:'''
        return self.forward_average_C_y(self.dim + 1)

    def forward_average_T_z(self):
        ''':meta private:'''
        return self.forward_average_C_z(self.dim + 1)

    def forward_average_S_y(self):
        ''':meta private:'''
        return self.forward_average_C_y(self.dim + 2)

    def forward_average_S_z(self):
        ''':meta private:'''
        return self.forward_average_C_z(self.dim + 2)

    @staticmethod
    def _backward_average_x(atom, i, j, k, x, y, z):
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # forward average
        atom[0] = dx * dy * dz / 2
        atom[1] = atom[0]

    def backward_average_v_y(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_average_x(atom[i, j, k, self.dim+1, 1, 1, :, 1], j, i, k, self.y, self.x, self.z)
        return atom

    def backward_average_w_z(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_average_x(atom[i, j, k, self.dim+1, 2, 1, 1, :], k, j, i, self.z, self.y, self.x)
        return atom

    def coriolis(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            # Value of yu at the position of v
            self._forward_average_x(atom[i, j, k, 1, 0, 0, :, 1], j, i, k, self.y, self.x, self.z)
            self._forward_average_x(atom[i, j, k, 1, 0, 1, :, 1], j, i, k, self.y, self.x, self.z)
            atom[i, j, k, 1, 0, :, :, :] *= self.y[j] / 2

            # Value of -yv at the position of u
            self._forward_average_x(atom[i, j, k, 0, 1, :, 0, 1], i, j, k, self.x, self.y, self.z)
            self._forward_average_x(atom[i, j, k, 0, 1, :, 1, 1], i, j, k, self.x, self.y, self.z)
            atom[i, j, k, 0, 1, :, :, :] *= -(self.y[j] + self.y[j-1]) / 4
        return atom

    def wind_stress(self):
        ''':meta private:'''
        alpha = self.get_parameter('Wind Stress Parameter')
        asym = self.get_parameter('Asymmetry Parameter')

        frc = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        for i, j, k in numpy.ndindex(self.nx - 1, self.ny, self.nz):
            # volume size in the x direction
            dx = (self.x[i+1] - self.x[i-1]) / 2
            # volume size in the y direction
            dy = self.y[j] - self.y[j-1]
            # volume size in the z direction
            dz = self.z[k] - self.z[k-1]

            y = (self.y[j] + self.y[j-1]) / 2
            frc[i, j, k, 0] = - (1 - asym) * numpy.cos(2 * numpy.pi * y) - asym * numpy.cos(numpy.pi * y)
            frc[i, j, k, 0] *= alpha / (2 * numpy.pi) * dx * dy * dz
        return utils.create_state_vec(frc, self.nx, self.ny, self.nz, self.dof)

    @staticmethod
    def _mass_x(atom, i, j, k, x, y, z):
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        atom[0] = dx * dy * dz

    def mass_x(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._mass_x(atom[i, j, k, 0:1], i, j, k, self.x, self.y, self.z)
        return atom

    def mass_y(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._mass_x(atom[i, j, k, 1:2], j, i, k, self.y, self.x, self.z)
        return atom

    def mass_z(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._mass_x(atom[i, j, k, 2:3], k, j, i, self.z, self.y, self.x)
        return atom

    @staticmethod
    def _mass_C(atom, i, j, k, x, y, z):
        # volume size in the x direction
        dx = x[i] - x[i-1]
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        atom[0] = dx * dy * dz

    def mass_C(self, var):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._mass_C(atom[i, j, k, var:var+1], i, j, k, self.x, self.y, self.z)
        return atom

    def mass_T(self):
        ''':meta private:'''
        return self.mass_C(self.dim + 1)

    def mass_S(self):
        ''':meta private:'''
        return self.mass_C(self.dim + 2)

    def average_x(self, state):
        ''':meta private:'''
        averages = numpy.zeros((self.nx+1, self.ny, self.nz))

        cropped_state = state[:, 1:self.ny+1, 1:self.nz+1]

        averages[:, :, :] += 1/2 * cropped_state[0:self.nx+1, :, :]
        averages[:, :, :] += 1/2 * cropped_state[1:self.nx+2, :, :]

        return averages

    def average_y(self, state):
        ''':meta private:'''
        averages = numpy.zeros((self.nx, self.ny+1, self.nz))

        cropped_state = state[1:self.nx+1, :, 1:self.nz+1]

        averages[:, :, :] += 1/2 * cropped_state[:, 0:self.ny+1, :]
        averages[:, :, :] += 1/2 * cropped_state[:, 1:self.ny+2, :]

        return averages

    def average_z(self, state):
        ''':meta private:'''
        averages = numpy.zeros((self.nx, self.ny, self.nz+1))

        cropped_state = state[1:self.nx+1, 1:self.ny+1, :]

        averages[:, :, :] += 1/2 * cropped_state[:, :, 0:self.nz+1]
        averages[:, :, :] += 1/2 * cropped_state[:, :, 1:self.nz+2]

        return averages

    @staticmethod
    def _weighted_average(atom, i, x):
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2

        # volume sizes associated with the v velocities
        dxmh = x[i] - x[i-1]
        dxph = x[i+1] - x[i]

        atom[0] = 1 / 2 * dxmh / dx
        atom[1] = 1 / 2 * dxph / dx

    def weighted_average_x(self, state):
        ''':meta private:'''
        averages = numpy.zeros((self.nx, self.ny+1, self.nz+1))

        cropped_state = state[:, 0:self.ny+1, 0:self.nz+1]

        atom = numpy.zeros(2)
        for i in range(self.nx):
            self._weighted_average(atom, i, self.x)

            averages[i, :, :] += atom[0] * cropped_state[i+1, :, :]
            averages[i, :, :] += atom[1] * cropped_state[i+2, :, :]

        return averages

    def weighted_average_y(self, state):
        ''':meta private:'''
        averages = numpy.zeros((self.nx+1, self.ny, self.nz+1))

        cropped_state = state[0:self.nx+1, :, 0:self.nz+1]

        atom = numpy.zeros(2)
        for j in range(self.ny):
            self._weighted_average(atom, j, self.y)

            averages[:, j, :] += atom[0] * cropped_state[:, j+1, :]
            averages[:, j, :] += atom[1] * cropped_state[:, j+2, :]

        return averages

    def weighted_average_z(self, state):
        ''':meta private:'''
        averages = numpy.zeros((self.nx+1, self.ny+1, self.nz))

        cropped_state = state[0:self.nx+1, 0:self.ny+1, :]

        atom = numpy.zeros(2)
        for k in range(self.nz):
            self._weighted_average(atom, k, self.z)

            averages[:, :, k] += atom[0] * cropped_state[:, :, k+1]
            averages[:, :, k] += atom[1] * cropped_state[:, :, k+2]

        return averages

    def u_u_x(self, atomJ, atomF, state):
        ''':meta private:'''
        averages = self.average_x(state[:, :, :, 0])

        atom = numpy.zeros(3)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._forward_u_x(atom, i, j, k, self.x, self.y, self.z)
            atomF[i, j, k, 0, 0, 0:2, 1, 1] -= atom[1] * averages[i, j, k] * 1 / 2
            atomF[i, j, k, 0, 0, 1:3, 1, 1] -= atom[2] * averages[i+1, j, k] * 1 / 2

            atomJ[i, j, k, 0, 0, 0:2, 1, 1] -= atom[1] * averages[i, j, k] * 1 / 2
            atomJ[i, j, k, 0, 0, 1:3, 1, 1] -= atom[2] * averages[i+1, j, k] * 1 / 2

    def u_v_x(self, atomJ, atomF, state):
        ''':meta private:'''
        averages_u = self.weighted_average_y(state[:, :, :, 0])
        averages_v = self.average_x(state[:, :, :, 1])

        atom = numpy.zeros(3)
        atom_average = numpy.zeros(2)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_y(atom, j, i, k, self.y, self.x, self.z)
            atomF[i, j, k, 1, 1, 0:2, 1, 1] -= atom[0] * averages_u[i, j, k+1] * 1 / 2
            atomF[i, j, k, 1, 1, 1:3, 1, 1] -= atom[1] * averages_u[i+1, j, k+1] * 1 / 2

            self._weighted_average(atom_average, j, self.y)
            atomJ[i, j, k, 1, 0, 0, 1:3, 1] -= atom[0] * averages_v[i, j, k] * atom_average
            atomJ[i, j, k, 1, 0, 1, 1:3, 1] -= atom[1] * averages_v[i+1, j, k] * atom_average

    def u_w_x(self, atomJ, atomF, state):
        ''':meta private:'''
        averages_u = self.weighted_average_z(state[:, :, :, 0])
        averages_w = self.average_x(state[:, :, :, 2])

        atom = numpy.zeros(3)
        atom_average = numpy.zeros(2)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_z(atom, k, j, i, self.z, self.y, self.x)
            atomF[i, j, k, 2, 2, 0:2, 1, 1] -= atom[0] * averages_u[i, j+1, k] * 1 / 2
            atomF[i, j, k, 2, 2, 1:3, 1, 1] -= atom[1] * averages_u[i+1, j+1, k] * 1 / 2

            self._weighted_average(atom_average, k, self.z)
            atomJ[i, j, k, 2, 0, 0, 1, 1:3] -= atom[0] * averages_w[i, j, k] * atom_average
            atomJ[i, j, k, 2, 0, 1, 1, 1:3] -= atom[1] * averages_w[i+1, j, k] * atom_average

    def u_C_x(self, atomJ, atomF, state, var):
        ''':meta private:'''
        averages_u = state[0:self.nx+1, 1:self.ny+1, 1:self.nz+1, 0]
        averages_C = self.average_x(state[:, :, :, var])

        atom = numpy.zeros(3)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_x(atom, i, j, k, self.x, self.y, self.z)
            atomF[i, j, k, var, var, 0:2, 1, 1] -= atom[0] * averages_u[i, j, k] * 1/2
            atomF[i, j, k, var, var, 1:3, 1, 1] -= atom[1] * averages_u[i+1, j, k] * 1/2

            atomJ[i, j, k, var, 0, 0, 1, 1] -= atom[0] * averages_C[i, j, k]
            atomJ[i, j, k, var, 0, 1, 1, 1] -= atom[1] * averages_C[i+1, j, k]

    def u_T_x(self, atomJ, atomF, state):
        ''':meta private:'''
        return self.u_C_x(atomJ, atomF, state, self.dim + 1)

    def u_S_x(self, atomJ, atomF, state):
        ''':meta private:'''
        return self.u_C_x(atomJ, atomF, state, self.dim + 2)

    def v_u_y(self, atomJ, atomF, state):
        ''':meta private:'''
        averages_u = self.average_y(state[:, :, :, 0])
        averages_v = self.weighted_average_x(state[:, :, :, 1])

        atom = numpy.zeros(3)
        atom_average = numpy.zeros(2)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_y(atom, i, j, k, self.x, self.y, self.z)
            atomF[i, j, k, 0, 0, 1, 0:2, 1] -= atom[0] * averages_v[i, j, k+1] * 1 / 2
            atomF[i, j, k, 0, 0, 1, 1:3, 1] -= atom[1] * averages_v[i, j+1, k+1] * 1 / 2

            self._weighted_average(atom_average, i, self.x)
            atomJ[i, j, k, 0, 1, 1:3, 0, 1] -= atom[0] * averages_u[i, j, k] * atom_average
            atomJ[i, j, k, 0, 1, 1:3, 1, 1] -= atom[1] * averages_u[i, j+1, k] * atom_average

    def v_v_y(self, atomJ, atomF, state):
        ''':meta private:'''
        averages = self.average_y(state[:, :, :, 1])

        atom = numpy.zeros(3)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._forward_u_x(atom, j, i, k, self.y, self.x, self.z)
            atomF[i, j, k, 1, 1, 1, 0:2, 1] -= atom[1] * averages[i, j, k] * 1 / 2
            atomF[i, j, k, 1, 1, 1, 1:3, 1] -= atom[2] * averages[i, j+1, k] * 1 / 2

            atomJ[i, j, k, 1, 1, 1, 0:2, 1] -= atom[1] * averages[i, j, k] * 1 / 2
            atomJ[i, j, k, 1, 1, 1, 1:3, 1] -= atom[2] * averages[i, j+1, k] * 1 / 2

    def v_w_y(self, atomJ, atomF, state):
        ''':meta private:'''
        averages_v = self.weighted_average_z(state[:, :, :, 1])
        averages_w = self.average_y(state[:, :, :, 2])

        atom = numpy.zeros(3)
        atom_average = numpy.zeros(2)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_y(atom, k, j, i, self.z, self.y, self.x)
            atomF[i, j, k, 2, 2, 1, 0:2, 1] -= atom[0] * averages_v[i+1, j, k] * 1 / 2
            atomF[i, j, k, 2, 2, 1, 1:3, 1] -= atom[1] * averages_v[i+1, j+1, k] * 1 / 2

            self._weighted_average(atom_average, k, self.z)
            atomJ[i, j, k, 2, 1, 1, 0, 1:3] -= atom[0] * averages_w[i, j, k] * atom_average
            atomJ[i, j, k, 2, 1, 1, 1, 1:3] -= atom[1] * averages_w[i, j+1, k] * atom_average

    def v_C_y(self, atomJ, atomF, state, var):
        ''':meta private:'''
        averages_v = state[1:self.nx+1, 0:self.ny+1, 1:self.nz+1, 1]
        averages_C = self.average_y(state[:, :, :, var])

        atom = numpy.zeros(3)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_x(atom, j, i, k, self.y, self.x, self.z)
            atomF[i, j, k, var, var, 1, 0:2, 1] -= atom[0] * averages_v[i, j, k] * 1/2
            atomF[i, j, k, var, var, 1, 1:3, 1] -= atom[1] * averages_v[i, j+1, k] * 1/2

            atomJ[i, j, k, var, 1, 1, 0, 1] -= atom[0] * averages_C[i, j, k]
            atomJ[i, j, k, var, 1, 1, 1, 1] -= atom[1] * averages_C[i, j+1, k]

    def v_T_y(self, atomJ, atomF, state):
        ''':meta private:'''
        return self.v_C_y(atomJ, atomF, state, self.dim + 1)

    def v_S_y(self, atomJ, atomF, state):
        ''':meta private:'''
        return self.v_C_y(atomJ, atomF, state, self.dim + 2)

    def w_u_z(self, atomJ, atomF, state):
        ''':meta private:'''
        averages_u = self.average_z(state[:, :, :, 0])
        averages_w = self.weighted_average_x(state[:, :, :, 2])

        atom = numpy.zeros(3)
        atom_average = numpy.zeros(2)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_z(atom, i, j, k, self.x, self.y, self.z)
            atomF[i, j, k, 0, 0, 1, 1, 0:2] -= atom[0] * averages_w[i, j+1, k] * 1 / 2
            atomF[i, j, k, 0, 0, 1, 1, 1:3] -= atom[1] * averages_w[i, j+1, k+1] * 1 / 2

            self._weighted_average(atom_average, i, self.x)
            atomJ[i, j, k, 0, 2, 1:3, 1, 0] -= atom[0] * averages_u[i, j, k] * atom_average
            atomJ[i, j, k, 0, 2, 1:3, 1, 1] -= atom[1] * averages_u[i, j, k+1] * atom_average

    def w_v_z(self, atomJ, atomF, state):
        ''':meta private:'''
        averages_v = self.average_z(state[:, :, :, 1])
        averages_w = self.weighted_average_y(state[:, :, :, 2])

        atom = numpy.zeros(3)
        atom_average = numpy.zeros(2)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_z(atom, j, i, k, self.y, self.x, self.z)
            atomF[i, j, k, 1, 1, 1, 1, 0:2] -= atom[0] * averages_w[i+1, j, k] * 1 / 2
            atomF[i, j, k, 1, 1, 1, 1, 1:3] -= atom[1] * averages_w[i+1, j, k+1] * 1 / 2

            self._weighted_average(atom_average, j, self.y)
            atomJ[i, j, k, 1, 2, 1, 1:3, 0] -= atom[0] * averages_v[i, j, k] * atom_average
            atomJ[i, j, k, 1, 2, 1, 1:3, 1] -= atom[1] * averages_v[i, j, k+1] * atom_average

    def w_w_z(self, atomJ, atomF, state):
        ''':meta private:'''
        averages = self.average_z(state[:, :, :, 2])

        atom = numpy.zeros(3)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._forward_u_x(atom, k, j, i, self.z, self.y, self.x)
            atomF[i, j, k, 2, 2, 1, 1, 0:2] -= atom[1] * averages[i, j, k] * 1 / 2
            atomF[i, j, k, 2, 2, 1, 1, 1:3] -= atom[2] * averages[i, j, k+1] * 1 / 2

            atomJ[i, j, k, 2, 2, 1, 1, 0:2] -= atom[1] * averages[i, j, k] * 1 / 2
            atomJ[i, j, k, 2, 2, 1, 1, 1:3] -= atom[2] * averages[i, j, k+1] * 1 / 2

    def w_C_z(self, atomJ, atomF, state, var):
        ''':meta private:'''
        averages_w = state[1:self.nx+1, 1:self.ny+1, 0:self.nz+1, 2]
        averages_C = self.average_z(state[:, :, :, var])

        atom = numpy.zeros(3)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_x(atom, k, j, i, self.z, self.y, self.x)
            atomF[i, j, k, var, var, 1, 1, 0:2] -= atom[0] * averages_w[i, j, k] * 1/2
            atomF[i, j, k, var, var, 1, 1, 1:3] -= atom[1] * averages_w[i, j, k+1] * 1/2

            atomJ[i, j, k, var, 2, 1, 1, 0] -= atom[0] * averages_C[i, j, k]
            atomJ[i, j, k, var, 2, 1, 1, 1] -= atom[1] * averages_C[i, j, k+1]

    def w_T_z(self, atomJ, atomF, state):
        ''':meta private:'''
        return self.w_C_z(atomJ, atomF, state, self.dim + 1)

    def w_S_z(self, atomJ, atomF, state):
        ''':meta private:'''
        return self.w_C_z(atomJ, atomF, state, self.dim + 2)
